#
# Copyright (c) 2021-2025, NVIDIA CORPORATION. All rights reserved.
#

# Standard Header Keys
HEADER_ACCEPT = "Accept"
HEADER_USER_AGENT = "User-Agent"
HEADER_CONTENT_TYPE = "Content-Type"
HEADER_CONTENT_LENGTH = "Content-Length"
HEADER_LOCATION = "Location"
# Standard Header Values
USER_AGENT_BASE = "ais/python"
JSON_CONTENT_TYPE = "application/json"
MSGPACK_CONTENT_TYPE = "application/msgpack"
# AIS Headers
AIS_CHECKSUM_TYPE = "ais-checksum-type"
AIS_CHECKSUM_VALUE = "ais-checksum-value"
AIS_ACCESS_TIME = "ais-atime"
AIS_VERSION = "ais-version"
AIS_CUSTOM_MD = "ais-custom-md"
AIS_BCK_NAME = "ais-bucket-name"
AIS_BCK_PROVIDER = "ais-bucket-provider"
AIS_OBJ_NAME = "ais-name"
AIS_LOCATION = "ais-location"
AIS_MIRROR_PATHS = "ais-mirror-paths"
AIS_MIRROR_COPIES = "ais-mirror-copies"
AIS_PRESENT = "ais-present"
# Bucket Props Header keys
HEADER_PREFIX = "ais-"
HEADER_BUCKET_PROPS = HEADER_PREFIX + "bucket-props"
HEADER_BUCKET_SUMM = HEADER_PREFIX + "bucket-summ"
HEADER_XACTION_ID = HEADER_PREFIX + "xaction-id"
HEADER_NODE_ID = HEADER_PREFIX + "node-id"
HEADER_NODE_URL = HEADER_PREFIX + "node-url"
# Object Props Header Keys
HEADER_OBJECT_BLOB_DOWNLOAD = HEADER_PREFIX + "blob-download"
HEADER_OBJECT_BLOB_CHUNK_SIZE = HEADER_PREFIX + "blob-chunk"
HEADER_OBJECT_BLOB_WORKERS = HEADER_PREFIX + "blob-workers"
HEADER_OBJECT_APPEND_HANDLE = "ais-append-handle"
# Ref: https://www.rfc-editor.org/rfc/rfc7233#section-2.1
HEADER_RANGE = "Range"
# AuthN Headers
HEADER_AUTHORIZATION = "Authorization"

# URL Params
# See api/apc/query.go
QPARAM_WHAT = "what"
QPARAM_PROVIDER = "provider"
QPARAM_BCK_TO = "bck_to"
QPARAM_FLT_PRESENCE = "presence"
QPARAM_BSUMM_REMOTE = "bsumm_remote"
QPARAM_KEEP_REMOTE = "keep_bck_md"
QPARAM_ARCHPATH = "archpath"
QPARAM_ARCHREGX = "archregx"
QPARAM_ARCHMODE = "archmode"
QPARAM_FORCE = "frc"
QPARAM_PRIMARY_READY_REB = "prr"
QPARAM_NAMESPACE = "namespace"
QPARAM_OBJ_APPEND = "append_type"
QPARAM_OBJ_APPEND_HANDLE = "append_handle"
DSORT_UUID = "uuid"
QPARAM_UUID = "uuid"
QPARAM_LATEST = "latest-ver"
QPARAM_NEW_CUSTOM = "set-new-custom"
# etl
QPARAM_ETL_NAME = "etl_name"
QPARAM_ETL_ARGS = "etl_args"
# etl websocket
ETL_WS_DESTINATION_ADDR = "dst_addr"
ETL_WS_FQN = "fqn"

# URL Param values
# See api/apc/query.go
WHAT_SMAP = "smap"
WHAT_ONE_XACT_STATUS = "status"
WHAT_ALL_XACT_STATUS = "status_all"
WHAT_ALL_RUNNING_STATUS = "running_all"
WHAT_QUERY_XACT_STATS = "qryxstats"
WHAT_NODE_STATS_AND_STATUS = "node_status"

# URL paths
URL_PATH_CLUSTER = "cluster"
URL_PATH_BUCKETS = "buckets"
URL_PATH_OBJECTS = "objects"
URL_PATH_HEALTH = "health"
URL_PATH_DAEMON = "daemon"
URL_PATH_ETL = "etl"
URL_PATH_DSORT = "sort"
URL_PATH_REVERSE = "reverse"
DSORT_ABORT = "abort"
# AuthN
URL_PATH_AUTHN_USERS = "users"
URL_PATH_AUTHN_CLUSTERS = "clusters"
URL_PATH_AUTHN_ROLES = "roles"
URL_PATH_AUTHN_TOKENS = "tokens"

# HTTP Methods
HTTP_METHOD_GET = "get"
HTTP_METHOD_POST = "post"
HTTP_METHOD_DELETE = "delete"
HTTP_METHOD_PUT = "put"
HTTP_METHOD_HEAD = "head"
HTTP_METHOD_PATCH = "patch"

# Actions
# See api/apc/actmsg.go
ACT_CREATE_BCK = "create-bck"
ACT_DESTROY_BCK = "destroy-bck"
ACT_COPY_BCK = "copy-bck"
ACT_ETL_BCK = "etl-bck"
ACT_EVICT_REMOTE_BCK = "evict-remote-bck"
ACT_LIST = "list"
ACT_MOVE_BCK = "move-bck"
ACT_PROMOTE = "promote"
ACT_SUMMARY_BCK = "summary-bck"
ACT_BLOB_DOWNLOAD = "blob-download"
# Multi-object actions
ACT_DELETE_OBJECTS = "delete-listrange"
ACT_EVICT_OBJECTS = "evict-listrange"
ACT_PREFETCH_OBJECTS = "prefetch-listrange"
ACT_COPY_OBJECTS = "copy-listrange"
ACT_TRANSFORM_OBJECTS = "etl-listrange"
ACT_ARCHIVE_OBJECTS = "archive"
# Job actions
ACT_START = "start"

# Defaults
DEFAULT_CHUNK_SIZE = 32768
DEFAULT_JOB_WAIT_TIMEOUT = 300
DEFAULT_DSORT_WAIT_TIMEOUT = 300
DEFAULT_DATASET_MAX_COUNT = 100000
DEFAULT_JOB_POLL_TIME = 0.2

# ENCODING
UTF_ENCODING = "utf-8"

# Status Codes
STATUS_ACCEPTED = 202
STATUS_NO_CONTENT = 204
STATUS_OK = 200
STATUS_BAD_REQUEST = 400
STATUS_PARTIAL_CONTENT = 206
STATUS_REDIRECT_TMP = 307
STATUS_REDIRECT_PERM = 301

# Protocol
HTTP = "http://"
HTTPS = "https://"

# Environment Variables
AIS_CLIENT_CA = "AIS_CLIENT_CA"
AIS_AUTHN_TOKEN = "AIS_AUTHN_TOKEN"
AIS_CLIENT_CRT = "AIS_CRT"
AIS_CLIENT_KEY = "AIS_CRT_KEY"

# Content Constants
LOREM = (
    "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod"
    " tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim"
    " veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea"
    " commodo consequat."
)
DUIS = (
    "Duis aute irure dolor in reprehenderit in voluptate velit esse cillum"
    " dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non"
    " proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
    " Et harum quidem.."
)

# AWS Constants
AWS_DEFAULT_REGION = "us-east-1"

# Time constants
NANOSECONDS_IN_SECOND = 1_000_000_000

DEFAULT_LOG_FORMAT = "%(asctime)s %(levelname)s: %(message)s"

# Ref: https://www.rfc-editor.org/rfc/rfc7233#section-2.1
BYTE_RANGE_PREFIX_LENGTH = 6

# Custom seed (MLCG32)
XX_HASH_SEED = 1103515245

# Job Masked Field Constants
# Ref:
# ┌──────────────────────────────────────────┬───────────────┬──────────────────────────┐
# │             bits 20 through 63           │ bits 10–19    │ bits 0–9                 │
# │        CHANNEL_BUFFERED_COUNT (44 bits)  │ WORKER_COUNT  │ JOGGER_COUNT             │
# └──────────────────────────────────────────┴───────────────┴──────────────────────────┘
JOGGER_COUNT_BITS = 10
WORKER_COUNT_BITS = 10
CHANNEL_COUNT_BITS = 64 - (JOGGER_COUNT_BITS + WORKER_COUNT_BITS)

# bit‐masks and shifts
JOGGER_COUNT_MASK = (1 << JOGGER_COUNT_BITS) - 1
WORKER_COUNT_SHIFT = JOGGER_COUNT_BITS
WORKER_COUNT_MASK = ((1 << WORKER_COUNT_BITS) - 1) << WORKER_COUNT_SHIFT
CHANNEL_COUNT_SHIFT = JOGGER_COUNT_BITS + WORKER_COUNT_BITS
CHANNEL_COUNT_MASK = ((1 << CHANNEL_COUNT_BITS) - 1) << CHANNEL_COUNT_SHIFT



#
# Copyright (c) 2022-2025, NVIDIA CORPORATION. All rights reserved.
#
from urllib.parse import urljoin, urlencode
from typing import TypeVar, Type, Any, Dict, Optional, Tuple, Union

from requests import Response
from tenacity import Retrying

from aistore.sdk.const import (
    JSON_CONTENT_TYPE,
    HEADER_USER_AGENT,
    USER_AGENT_BASE,
    HEADER_CONTENT_TYPE,
    HEADER_AUTHORIZATION,
    HTTPS,
    HEADER_LOCATION,
    STATUS_REDIRECT_PERM,
    STATUS_REDIRECT_TMP,
    URL_PATH_DAEMON,
    WHAT_SMAP,
    QPARAM_WHAT,
    HTTP_METHOD_GET,
)
from aistore.sdk.response_handler import ResponseHandler, AISResponseHandler
from aistore.sdk.session_manager import SessionManager
from aistore.version import __version__ as sdk_version
from aistore.sdk.types import Smap
from aistore.sdk.utils import decode_response, get_logger
from aistore.sdk.retry_config import RetryConfig

T = TypeVar("T")

logger = get_logger(__name__)


class RequestClient:
    """
    Internal client for making requests to an AIS cluster or a specific proxy (e.g. AuthN).

    Args:
        endpoint (str): Endpoint to either the AIS cluster or a specific proxy (e.g. AuthN).
        session_manager (SessionManager): SessionManager for creating and accessing requests session.
        timeout (Union[float, Tuple[float, float], None], optional): Request timeout in seconds; a single float.
            for both connect/read timeouts (e.g., 5.0), a tuple for separate connect/read timeouts (e.g., (3.0, 10.0)),
            or None to disable timeout.
        token (str, optional): Authorization token.
        response_handler (ResponseHandler): Handler for processing HTTP responses. Defaults to AISResponseHandler.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        endpoint: str,
        session_manager: SessionManager,
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        token: str = None,
        response_handler: ResponseHandler = AISResponseHandler(),
        network_retry_config: Retrying = None,
    ):
        self._base_url = urljoin(endpoint, "v1")
        self._session_manager = session_manager
        self._token = token
        self._timeout = timeout
        self._response_handler = response_handler
        # smap is used to calculate the target node for a given object
        self._smap = None
        self._network_retry_config = (
            network_retry_config or RetryConfig.default().network_retry
        )

    @property
    def base_url(self):
        """Return the base URL."""
        return self._base_url

    @property
    def timeout(self):
        """Return the timeout for requests."""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: Union[float, Tuple[float, float]]):
        """
        Set the timeout for all requests from this client.

        Args:
            timeout: Timeout for requests.
        """
        self._timeout = timeout

    @property
    def session_manager(self) -> SessionManager:
        """
        Return the SessionManager used to create sessions for this client.
        """
        return self._session_manager

    @session_manager.setter
    def session_manager(self, session_manager):
        self._session_manager = session_manager

    @property
    def token(self) -> str:
        """
        Return the token for authorization.
        """
        return self._token

    @property
    def network_retry_config(self) -> Retrying:
        """
        Return the network retry configuration for this client.
        """
        return self._network_retry_config

    @token.setter
    def token(self, token: str):
        """
        Set the token for Authorization.

        Args:
            token (str): Token for authorization.
        """
        self._token = token

    def get_smap(self, force_update: bool = False) -> "Smap":
        """Return the smap."""
        if not self._smap or force_update:
            self._smap = self.request_deserialize(
                HTTP_METHOD_GET,
                path=URL_PATH_DAEMON,
                res_model=Smap,
                params={QPARAM_WHAT: WHAT_SMAP},
            )
        return self._smap

    def clone(self, base_url: Optional[str] = None) -> "RequestClient":
        """
        Create a copy of the current RequestClient instance with an optional new base URL.

        Args:
            base_url (Optional[str]): New base URL for the cloned client. Defaults to the existing base URL.

        Returns:
            RequestClient: A new instance with the same settings but an optional different base URL.
        """

        # Default to the existing base URL if none is provided
        base_url = base_url or self._base_url

        # Ensure the base URL ends with "/v1"
        base_url = base_url if base_url.endswith("/v1") else urljoin(base_url, "v1")

        return RequestClient(
            endpoint=base_url,
            session_manager=self._session_manager,
            timeout=self._timeout,
            token=self._token,
            response_handler=self._response_handler,
        )

    def request_deserialize(
        self, method: str, path: str, res_model: Type[T], **kwargs
    ) -> T:
        """
        Make a request and deserialize the response to a defined type.

        Args:
            method (str): HTTP method (e.g. POST, GET, PUT, DELETE).
            path (str): URL path to call.
            res_model (Type[T]): Resulting type to which the response should be deserialized.
            **kwargs (optional): Optional keyword arguments to pass with the call to request.

        Returns:
            Parsed result of the call to the API, as res_model.
        """
        resp = self.request(method, path, **kwargs)
        return decode_response(res_model, resp)

    def _retryable_session_request(
        self, method: str, url: str, headers: dict, **kwargs
    ) -> Response:
        """
        Makes an HTTP request with automatic retries on connection-related errors.

        Retries the request if it encounters transient network issues such as:
        - ConnectTimeout
        - ReadTimeout
        - ConnectionError (refused connection, reset by peer)
        - AISRetryableError (custom error for AIS)
        - ChunkedEncodingError

        If the request is an HTTPS request with a payload (`data`), it uses `_request_with_manual_redirect`.
        Otherwise, it falls back to `_session_request`.

        **Why `tenacity.retry` over `urllib3.Retry` for this function?**
        `urllib3.Retry` always retries the **same failing URL**, which is problematic if a target is down or restarting.
        Instead, we retry via the **proxy URL** to reach another available target—something `urllib3.Retry` does not
        support.

        Args:
            method (str): HTTP method (e.g., GET, POST).
            url (str): Target URL.
            headers (dict): HTTP headers.
            kwargs: Additional request parameters.

        Returns:
            Response: The HTTP response from the server.
        """

        if url.startswith(HTTPS) and "data" in kwargs:
            response = self._request_with_manual_redirect(
                method, url, headers, **kwargs
            )
        else:
            response = self._session_request(method, url, headers, **kwargs)
        return self._response_handler.handle_response(response)

    def request(
        self,
        method: str,
        path: str,
        endpoint: str = None,
        headers: Dict = None,
        **kwargs,
    ) -> Response:
        """
        Make a request.

        Args:
            method (str): HTTP method (e.g. POST, GET, PUT, DELETE).
            path (str): URL path to call.
            endpoint (str): Alternative endpoint for the AIS cluster (e.g. for connecting to a specific proxy).
            headers (dict): Extra headers to be passed with the request. Content-Type and User-Agent will be overridden.
            **kwargs (optional): Optional keyword arguments to pass with the call to request.

        Returns:
            Raw response from the API.
        """
        base = urljoin(endpoint, "v1") if endpoint else self._base_url
        url = f"{base}/{path.lstrip('/')}"
        if headers is None:
            headers = {}
        headers[HEADER_CONTENT_TYPE] = JSON_CONTENT_TYPE
        headers[HEADER_USER_AGENT] = f"{USER_AGENT_BASE}/{sdk_version}"
        if self.token:
            headers[HEADER_AUTHORIZATION] = f"Bearer {self.token}"
        # If the request is a GET to the daemon, we need to set the User-Agent header to avoid a 403 Forbidden
        return self.network_retry_config(
            self._retryable_session_request, method, url, headers, **kwargs
        )

    def _request_with_manual_redirect(
        self, method: str, url: str, headers, **kwargs
    ) -> Response:
        """
        Make a request to the proxy, close the session, and use a new session to make a request to the redirected
        target.

        This exists because the current implementation of `requests` does not seem to handle a 307 redirect
        properly from a server with TLS enabled with data in the request, and will error with the following on the
        initial connection to the proxy:
            SSLEOFError(8, 'EOF occurred in violation of protocol (_ssl.c:2406)')
        Instead, this implementation will not send the data to the proxy, and only use it to access the proper target.

        Args:
            method (str): HTTP method (e.g. POST, GET, PUT, DELETE).
            url (str): Initial AIS url.
            headers (dict): Extra headers to be passed with the request. Content-Type and User-Agent will be overridden.
            **kwargs (optional): Optional keyword arguments to pass with the call to request.

        Returns:
            Final response from AIS target

        """
        # Do not include data payload in the initial request to the proxy
        proxy_request_kwargs = {
            "headers": headers,
            "allow_redirects": False,
            **{k: v for k, v in kwargs.items() if k != "data"},
        }

        # Request to proxy, which should redirect
        resp = self.session_manager.session.request(method, url, **proxy_request_kwargs)
        if resp.status_code in (STATUS_REDIRECT_PERM, STATUS_REDIRECT_TMP):
            target_url = resp.headers.get(HEADER_LOCATION)
            # Redirected request to target
            resp = self._session_request(method, target_url, headers, **kwargs)
        return resp

    def _session_request(self, method, url, headers, **kwargs) -> Response:
        request_kwargs = {"headers": headers, **kwargs}
        if self._timeout is not None and "timeout" not in request_kwargs:
            request_kwargs["timeout"] = self._timeout
        return self.session_manager.session.request(method, url, **request_kwargs)

    def get_full_url(self, path: str, params: Dict[str, Any]) -> str:
        """
        Get the full URL to the path on the cluster with the given parameters.

        Args:
            path (str): Path on the cluster.
            params (Dict[str, Any]): Query parameters to include.

        Returns:
            URL including cluster base URL and parameters.

        """
        return f"{self._base_url}/{path.lstrip('/')}?{urlencode(params)}"



#
# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations
from typing import Optional, Union, Dict, List
from pydantic.v1 import BaseModel, Field
from aistore.sdk.authn.access_attr import AccessAttr
from aistore.sdk.const import NANOSECONDS_IN_SECOND
from aistore.sdk.types import BucketModel


# pylint: disable=missing-class-docstring, missing-function-docstring


class LoginMsg(BaseModel):
    """
    Represents a login message with a password and optional expiration duration.

    Attributes:
        password (str): The password string.
        expires_in (Optional[Union[int, float]]): The expiration duration in seconds.
    """

    password: str
    expires_in: Optional[Union[int, float]] = Field(
        None, description="Expiration duration in seconds."
    )

    def as_dict(self):
        """
        Converts the instance to a dict, converting the expiration duration to nanoseconds if specified.

        Returns:
            Dict[str, Union[str, int]]: The dict representation of the login message.
        """

        data = self.dict()
        if self.expires_in is not None:
            data["expires_in"] = int(
                self.expires_in * NANOSECONDS_IN_SECOND
            )  # Convert seconds to nanoseconds
        return data


class TokenMsg(BaseModel):
    """
    Represents a message containing a token.

    Attributes:
        token (str): The token string.
    """

    token: str


class ClusterInfo(BaseModel):
    """
    Represents information about a cluster.

    Attributes:
        id (str): The unique identifier of the cluster.
        alias (Optional[str]): The alias name of the cluster. Defaults to None.
        urls (List[str]): A list of URLs associated with the cluster.
    """

    id: str
    alias: Optional[str] = None
    urls: List[str] = []


class ClusterList(BaseModel):
    """
    Represents a list of clusters.

    Attributes:
        clusters (Dict[str, ClusterInfo]): A dictionary of cluster IDs to ClusterInfo objects.
    """

    clusters: Dict[str, ClusterInfo] = {}


class ClusterPermission(BaseModel):
    """
    Represents a cluster with its associated permissions.
    """

    id: str
    perm: str

    def describe(self) -> str:
        return AccessAttr.describe(int(self.perm))

    def __str__(self) -> str:
        return f"ClusterPermission(id={self.id}, perm={self.describe()})"


class BucketPermission(BaseModel):
    """
    Represents a bucket with its associated permissions.
    """

    bck: BucketModel
    perm: str

    def describe(self) -> str:
        return AccessAttr.describe(int(self.perm))

    def __str__(self) -> str:
        return f"BucketPermission(bck={self.bck}, perm={self.describe()})"


class RoleInfo(BaseModel):
    """
    Represents role information including permissions for clusters and buckets.

    Attributes:
        name (str): Name of the role.
        desc (str): Description of the role.
        clusters (Optional[List[ClusterPermission]]): List of cluster permissions.
        buckets (Optional[List[BucketPermission]]): List of bucket permissions.
        admin (bool): Whether the role has admin privileges.
    """

    name: str
    desc: str
    clusters: List[ClusterPermission] = None
    buckets: List[BucketPermission] = None
    admin: bool = False

    def __str__(self) -> str:
        if self.clusters is None:
            clusters_str = "None"
        else:
            clusters_str = ", ".join(str(cluster) for cluster in self.clusters)

        if self.buckets is None:
            buckets_str = "None"
        else:
            buckets_str = ", ".join(str(bucket) for bucket in self.buckets)

        return (
            f"RoleInfo(name={self.name}, desc={self.desc}, "
            f"clusters=[{clusters_str}], buckets=[{buckets_str}], admin={self.admin})"
        )


class RolesList(BaseModel):
    """
    Represents a list of roles.

    Attributes:
        __root__ (List[RoleInfo]): List of roles.
    """

    __root__: List[RoleInfo]

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]

    def __len__(self):
        return len(self.__root__)

    @property
    def roles(self):
        return self.__root__

    def __str__(self) -> str:
        return "\n".join(str(role) for role in self.__root__)


class UserInfo(BaseModel):
    """
    Represents user information in the AuthN service.

    Attributes:
        id (str): The username or ID of the user.
        password (str, optional): The user's password. Serialized as 'pass' in the request.
        roles (RolesList): The list of roles assigned to the user.
    """

    id: str
    password: Optional[str] = None
    roles: RolesList

    def dict(self, **kwargs):
        """
        Override the dict method to serialize the 'password' field as 'pass'.

        Returns:
            Dict[str, Union[str, RolesList]]: The dict representation of the user information.
        """
        user_dict = super().dict(**kwargs)
        if "password" in user_dict and user_dict["password"] is not None:
            user_dict["pass"] = user_dict.pop("password")
        return user_dict


class UsersList(BaseModel):
    """
    Represents a list of users.

    Attributes:
        __root__ (Dict[str, UserInfo]): Dictionary of user names/IDs to UserInfo objects.
    """

    __root__: Dict[str, UserInfo]

    def __iter__(self):
        return iter(self.__root__.values())

    def __getitem__(self, item):
        return self.__root__[item]

    def __len__(self):
        return len(self.__root__)

    @property
    def users(self):
        return self.__root__

    def __str__(self) -> str:
        return "\n".join(str(user) for user in self.__root__)



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

import os
import logging
from abc import ABC, abstractmethod
import sys


class ETLServer(ABC):
    """
    Abstract base class for all ETL servers.

    Provides:
    - `host_target`: the AIS target URL from environment variable `AIS_TARGET_URL`.
    - `logger`: a class-specific logger with INFO level (can be adjusted).
    """

    def __init__(self):
        """
        Initialize the base ETL server.

        Sets up:
        - `host_target`: AIS target URL from the `AIS_TARGET_URL` environment variable.
        - `arg_type`: Determines how to interpret the input path ("fqn", "url", or "").
        - `logger`: A class-specific logger configured to output to stdout.

        Subclasses can extend this method to initialize any transformation-specific
        resources (e.g., preloaded models, hash functions, lookup tables) required
        by the `transform()` method.

        Note:
            If you override `__init__` in a subclass, make sure to call
            `super().__init__()` to ensure proper base class initialization.
        """
        self.host_target = os.getenv("AIS_TARGET_URL")
        if not self.host_target:
            raise EnvironmentError("Environment variable 'AIS_TARGET_URL' must be set.")
        self.arg_type = os.getenv("ARG_TYPE", "").lower()  # "", "fqn" or "url"
        self.direct_put = os.getenv("DIRECT_PUT", "false").lower() == "true"

        # Configure logging
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            stream=sys.stdout,
        )
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

    @abstractmethod
    def transform(self, data: bytes, path: str, etl_args: str) -> bytes:
        """
        Transform the data received from a request.

        Args:
            data (bytes): The original content fetched from AIS.
            path (str): The request path. This is usually the path
                to the object.
            etl_args (str): The arguments passed for the transformation.

        Returns:
            bytes: Transformed data to return to the caller.
        """

    @abstractmethod
    def start(self):
        """
        Start the ETL server (blocking call).
        Typically binds and listens on a port.
        """

    def get_mime_type(self) -> str:
        """
        Optional override to specify MIME type of transformed response.

        Args:
            path (str): Request path or object path.

        Returns:
            str: MIME type (e.g., "application/json", "text/plain").
        """
        return "application/octet-stream"



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#

import os
from http.server import HTTPServer, BaseHTTPRequestHandler
from socketserver import ThreadingMixIn
from typing import Type, Tuple
import signal
import threading
from urllib.parse import unquote, urlparse, parse_qs

import requests

from aistore.sdk.etl.webserver.base_etl_server import ETLServer
from aistore.sdk.utils import compose_etl_direct_put_url
from aistore.sdk.const import (
    HEADER_CONTENT_LENGTH,
    HEADER_CONTENT_TYPE,
    HEADER_NODE_URL,
    STATUS_OK,
    STATUS_NO_CONTENT,
    QPARAM_ETL_ARGS,
)


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """
    Multi-threaded HTTP server that delegates ETL logic to a provided ETLServer instance.
    """

    def __init__(
        self,
        server_address: Tuple[str, int],
        handler_class: Type[BaseHTTPRequestHandler],
        etl_server: ETLServer,
    ):
        super().__init__(server_address, handler_class)
        self.etl_server = etl_server


class HTTPMultiThreadedServer(ETLServer):
    """
    Multi-threaded HTTP server implementation for ETL payload transformation.
    Handles GET and PUT requests via a request handler class.
    """

    def __init__(self, host: str = "0.0.0.0", port: int = 80):
        super().__init__()
        self.host = host
        self.port = port

    # pylint: disable=invalid-name
    class RequestHandler(BaseHTTPRequestHandler):
        """
        Request handler for the ETL HTTP server.
        Forwards GET/PUT requests to the AIS target and applies transformation logic.
        """

        def _set_headers(self, status_code: int = STATUS_OK, length: int = 0):
            self.send_response(status_code)
            mime_type = self.server.etl_server.get_mime_type()
            self.send_header(HEADER_CONTENT_TYPE, mime_type)
            self.send_header(HEADER_CONTENT_LENGTH, str(length))
            self.end_headers()

        def log_request(self, code="-", size="-"):
            # Suppress default request logging (or override as needed)
            pass

        def _direct_put(self, direct_put_url: str, data: bytes) -> bool:
            """
            Sends the transformed object directly to the specified AIS node (`direct_put_url`),
            eliminating the additional network hop through the original target.
            Used only in bucket-to-bucket offline transforms.
            """
            logger = self.server.etl_server.logger
            try:
                url = compose_etl_direct_put_url(
                    direct_put_url, self.server.etl_server.host_target
                )
                resp = requests.put(url, data, timeout=None)
                if resp.status_code == 200:
                    return True

                error = resp.text()
                logger.error(
                    "Failed to deliver object to %s: HTTP %s, %s",
                    direct_put_url,
                    resp.status_code,
                    error,
                )
            except Exception as e:
                logger.error("Exception during delivery to %s: %s", direct_put_url, e)

            return False

        def _get_fqn_content(self, path: str) -> bytes:
            """
            Parses and safely reads a file when using arg_type == 'fqn'.
            """
            decoded_path = unquote(path)
            safe_path = os.path.normpath(os.path.join("/", decoded_path.lstrip("/")))

            self.server.etl_server.logger.debug("Reading local file: %s", safe_path)

            with open(safe_path, "rb") as f:
                return f.read()

        def do_GET(self):
            """
            Handle GET requests by forwarding them to the AIS target or reading from FQN,
            applying the ETL transformation and returning the result.
            """

            logger = self.server.etl_server.logger

            parsed = urlparse(self.path)
            raw_path = parsed.path
            params = parse_qs(parsed.query)
            etl_args = params.get(QPARAM_ETL_ARGS, [""])[0]

            logger.debug("Received GET request for path: %s", raw_path)

            # Health check
            if raw_path == "/health":
                resp = b"Running"
                self._set_headers(length=len(resp))
                self.wfile.write(resp)
                return

            try:
                if self.server.etl_server.arg_type == "fqn":
                    content = self._get_fqn_content(raw_path)
                else:
                    target_url = f"{self.server.etl_server.host_target}{raw_path}"
                    logger.debug("Forwarding GET to AIS target: %s", target_url)

                    resp = requests.get(target_url, timeout=None)

                    if resp.status_code != STATUS_OK:
                        logger.warning(
                            "Failed to fetch from target %s (status %d)",
                            target_url,
                            resp.status_code,
                        )
                        self.send_error(
                            resp.status_code, "Failed to retrieve data from target"
                        )
                        return

                    content = resp.content

                transformed = self.server.etl_server.transform(
                    content, raw_path, etl_args
                )

                direct_put_url = self.headers.get(HEADER_NODE_URL)
                if direct_put_url:
                    success = self._direct_put(direct_put_url, transformed)
                    if success:
                        self._set_headers(status_code=STATUS_NO_CONTENT, length=0)
                        return

                self._set_headers(length=len(transformed))
                self.wfile.write(transformed)

            except FileNotFoundError:
                logger.error("File not found: %s", raw_path)
                self.send_error(404, f"Local file not found: {raw_path}")

            except requests.RequestException as e:
                logger.error("Request to AIS target failed: %s", str(e))
                self.send_error(500, f"Error contacting AIS target: {e}")

            except Exception as e:
                logger.exception("Unexpected error processing GET request")
                self.send_error(500, f"Unhandled error: {e}")

        def do_PUT(self):
            """
            Handle PUT requests by transforming the incoming data and responding with the transformed data.
            """
            logger = self.server.etl_server.logger
            parsed = urlparse(self.path)
            raw_path = parsed.path
            params = parse_qs(parsed.query)
            etl_args = params.get(QPARAM_ETL_ARGS, [""])[0]

            logger.debug("Received PUT request for path: %s", raw_path)

            try:
                if self.server.etl_server.arg_type == "fqn":
                    content = self._get_fqn_content(raw_path)
                else:
                    content_length = int(self.headers.get(HEADER_CONTENT_LENGTH, 0))
                    content = self.rfile.read(content_length)
                transformed = self.server.etl_server.transform(
                    content, raw_path, etl_args
                )

                direct_put_url = self.headers.get(HEADER_NODE_URL)
                if direct_put_url:
                    success = self._direct_put(direct_put_url, transformed)
                    if success:
                        self._set_headers(status_code=STATUS_NO_CONTENT, length=0)
                        return

                self._set_headers(length=len(transformed))
                self.wfile.write(transformed)

            except FileNotFoundError:
                logger.error("File not found: %s", raw_path)
                self.send_error(404, f"Local file not found: {raw_path}")

            except Exception as e:
                logger.error("Error processing PUT request: %s", str(e))
                self.send_error(500, "Internal error during transformation")

    def start(self):
        """
        Starts the HTTP server and gracefully handles SIGTERM/SIGINT for Kubernetes.
        """
        server = ThreadedHTTPServer(
            (self.host, self.port), self.RequestHandler, etl_server=self
        )
        self.logger.info(
            "HTTP Multi-threaded server running at %s:%s", self.host, self.port
        )

        # Register shutdown handler
        def shutdown_handler(signum, _frame):
            self.logger.info("Received signal %s: shutting down.", signum)
            threading.Thread(target=server.shutdown).start()

        signal.signal(signal.SIGTERM, shutdown_handler)
        signal.signal(signal.SIGINT, shutdown_handler)

        try:
            server.serve_forever()
        finally:
            server.server_close()
            self.logger.info("Server shutdown complete.")



#
# Copyright (c) 2021-2025, NVIDIA CORPORATION. All rights reserved.
#
from __future__ import annotations
import base64
from typing import Any, Mapping, List, Optional, Dict
import msgspec
from pydantic.v1 import BaseModel, Field, validator
from requests.structures import CaseInsensitiveDict

from aistore.sdk.errors import NoTargetError
from aistore.sdk.namespace import Namespace
from aistore.sdk.list_object_flag import ListObjectFlag
from aistore.sdk.obj.object_props import ObjectProps
from aistore.sdk.const import (
    HEADER_CONTENT_LENGTH,
    AIS_CHECKSUM_VALUE,
    AIS_ACCESS_TIME,
    AIS_VERSION,
    AIS_OBJ_NAME,
    AIS_LOCATION,
    AIS_MIRROR_COPIES,
    JOGGER_COUNT_MASK,
    WORKER_COUNT_MASK,
    WORKER_COUNT_SHIFT,
    CHANNEL_COUNT_MASK,
    CHANNEL_COUNT_SHIFT,
)
from aistore.sdk.utils import get_digest, xoshiro256_hash


# pylint: disable=too-few-public-methods,unused-variable,missing-function-docstring,too-many-lines
class ActionMsg(BaseModel):
    """
    Represents the action message passed by the client via json
    """

    action: str
    name: str = ""
    value: Any = None


class NetInfo(BaseModel):
    """
    Represents a set of network-related info
    """

    node_hostname: str = ""
    daemon_port: str = ""
    direct_url: str = ""


class Snode(BaseModel):
    """
    Represents a system node
    """

    daemon_id: str
    daemon_type: str
    public_net: NetInfo = None
    intra_control_net: NetInfo = None
    intra_data_net: NetInfo = None
    flags: int = 0
    id_digest: int = 0

    def in_maint_or_decomm(self) -> bool:
        return (self.flags & (1 << 2 | 1 << 3)) != 0


class Smap(BaseModel):
    """
    Represents a system map
    """

    tmap: Mapping[str, Snode]
    pmap: Mapping[str, Snode]
    proxy_si: Snode
    version: int = 0
    uuid: str = ""
    creation_time: str = ""

    def get_target_for_object(self, uname: str) -> Snode:
        """
        Determine the target node responsible for an object based on its bucket path and name.

        Args:
            uname (str): Fully qualified (namespaced) object name (e.g., f"{bck.get_path()}{obj.name}").

        Returns:
            Snode: The assigned target node.

        Raises:
            AISError: If no suitable target node is found.
        """
        digest = get_digest(uname)

        selected_node, max_hash = None, -1

        for tsi in self.tmap.values():
            if tsi.in_maint_or_decomm():
                continue  # Skip nodes in maintenance or decommissioned mode

            # Compute hash using Xoshiro256
            cs = xoshiro256_hash(tsi.id_digest ^ digest)

            if cs > max_hash:
                max_hash, selected_node = cs, tsi

        if selected_node is None:
            raise NoTargetError(len(self.tmap))

        return selected_node


class BucketEntry(msgspec.Struct):
    """
    Represents a single entry in a bucket -- an object
    See cmn/objlist.go/LsoEntry
    """

    n: str
    cs: str = ""
    a: str = ""
    v: str = ""
    t: str = ""
    s: int = 0
    c: int = 0
    f: int = 0
    object: Any = None

    def generate_object_props(self) -> ObjectProps:
        """
        Convert bucket entry data into Object Props.

        NOTE: Bucket entry data and object props are not a one-to-one mapping.

        Returns:
            ObjectProps with object data
        """
        headers = CaseInsensitiveDict(
            {
                AIS_OBJ_NAME: self.name,
                AIS_CHECKSUM_VALUE: self.checksum,
                AIS_ACCESS_TIME: self.atime,
                AIS_VERSION: self.version,
                AIS_LOCATION: self.location,
                HEADER_CONTENT_LENGTH: self.size,
                AIS_MIRROR_COPIES: self.copies,
            }
        )
        return ObjectProps(headers)

    @property
    def name(self):
        return self.n

    @property
    def checksum(self):
        return self.cs

    @property
    def atime(self):
        return self.a

    @property
    def version(self):
        return self.v

    @property
    def location(self):
        return self.t

    @property
    def size(self):
        return self.s

    @property
    def copies(self):
        return self.c

    @property
    def flags(self):
        return self.f

    def is_cached(self):
        return (self.flags & (1 << 6)) != 0

    def is_ok(self):
        return (self.flags & ((1 << 5) - 1)) == 0


class BucketList(msgspec.Struct):
    """
    Represents the response when getting a list of bucket items, containing a list of BucketEntry objects
    """

    UUID: str
    ContinuationToken: str
    Flags: int
    Entries: List[BucketEntry] = None

    @property
    def uuid(self):
        return self.UUID

    @property
    def continuation_token(self):
        return self.ContinuationToken

    @property
    def flags(self):
        return self.Flags

    @property
    def entries(self):
        return [] if self.Entries is None else self.Entries

    def get_entries(self):
        """
        Deprecated -- use entries property
        """
        return self.entries


class BucketModel(BaseModel):
    """
    Represents the response from the API containing bucket info
    """

    name: str
    provider: str
    namespace: Namespace = None

    def as_dict(self):
        dict_rep = {"name": self.name, "provider": self.provider}
        if self.namespace:
            dict_rep["namespace"] = self.namespace
        return dict_rep


class BsummCtrlMsg(BaseModel):
    """
    Represents the bucket summary control message
    """

    uuid: str
    prefix: str
    cached: bool
    present: bool


class JobArgs(BaseModel):
    """
    Represents the set of args to pass when making a job-related request
    """

    id: str = ""
    kind: str = ""
    daemon_id: str = ""
    bucket: BucketModel = None
    buckets: List[BucketModel] = None

    def as_dict(self):
        return {
            "ID": self.id,
            "Kind": self.kind,
            "DaemonID": self.daemon_id,
            "Bck": self.bucket,
            "Buckets": self.buckets,
        }


class JobQuery(BaseModel):
    """
    Structure to send the API when querying the cluster for multiple jobs
    """

    active: bool = False
    kind: str = ""
    target: str = ""

    def as_dict(self):
        return {
            "kind": self.kind,
            "node": self.target,
            "show_active": self.active,
        }


class JobStatus(BaseModel):
    """
    Represents the response of an API query to fetch job status
    """

    uuid: str = ""
    err: str = ""
    end_time: int = 0
    aborted: bool = False


class ETLInfo(BaseModel):  # pylint: disable=too-few-public-methods,unused-variable
    """
    Represents the API response when querying an ETL
    """

    id: str = ""
    xaction_id: str = ""
    stage: str = ""
    obj_count: int = 0
    in_bytes: int = 0
    out_bytes: int = 0


class ETLDetails(BaseModel):
    """
    Represents the API response of queries on single ETL details
    """

    id: str
    communication: str
    timeout: str
    code: Optional[bytes]
    spec: Optional[str]
    dependencies: Optional[str]
    runtime: Optional[str]  # see ext/etl/runtime/all.go
    chunk_size: int = 0
    argument: str = ""

    @validator("code")
    def set_code(cls, code):  # pylint: disable=no-self-argument
        if code is not None:
            code = base64.b64decode(code)
        return code

    @validator("spec")
    def set_spec(cls, spec):  # pylint: disable=no-self-argument
        if spec is not None:
            spec = base64.b64decode(spec)
        return spec


class InitETLArgs(BaseModel):
    """
    Represents the args shared by ETL initialization with code or spec
    """

    etl_name: str
    communication_type: str
    timeout: str
    arg_type: str = ""


class InitSpecETLArgs(InitETLArgs):
    """
    Represents the set of args the sdk will pass to AIStore when making a request to initialize an ETL with a spec
    """

    spec: str

    def as_dict(self):
        return {
            "id": self.etl_name,
            "timeout": self.timeout,
            "communication": f"{self.communication_type}://",
            "spec": self.spec,
            "argument": self.arg_type,
        }


class InitCodeETLArgs(InitETLArgs):
    """
    Represents the set of args the sdk will pass to AIStore when making a request to initialize an ETL with code
    """

    runtime: str
    dependencies: str
    functions: Dict[str, str]
    code: str
    chunk_size: int = None

    def as_dict(self):
        dict_rep = {
            "id": self.etl_name,
            "runtime": self.runtime,
            "communication": f"{self.communication_type}://",
            "timeout": self.timeout,
            "funcs": self.functions,
            "code": self.code,
            "dependencies": self.dependencies,
            "argument": self.arg_type,
        }
        if self.chunk_size:
            dict_rep["chunk_size"] = self.chunk_size
        return dict_rep


class PromoteAPIArgs(BaseModel):
    """
    Represents the set of args the sdk will pass to AIStore when making a promote request and
    provides conversion to the expected json format
    """

    target_id: str = ""
    source_path: str = ""
    object_name: str = ""
    recursive: bool = False
    overwrite_dest: bool = False
    delete_source: bool = False
    src_not_file_share: bool = False

    def as_dict(self):
        return {
            "tid": self.target_id,
            "src": self.source_path,
            "obj": self.object_name,
            "rcr": self.recursive,
            "ovw": self.overwrite_dest,
            "dls": self.delete_source,
            "notshr": self.src_not_file_share,
        }


class JobStats(BaseModel):
    """
    Represents performance statistics for a job.
    """

    objects: int = Field(default=0, alias="loc-objs")
    bytes: int = Field(default=0, alias="loc-bytes")
    out_objects: int = Field(default=0, alias="out-objs")
    out_bytes: int = Field(default=0, alias="out-bytes")
    in_objects: int = Field(default=0, alias="in-objs")
    in_bytes: int = Field(default=0, alias="in-bytes")

    class Config:  # pylint: disable=missing-class-docstring
        allow_population_by_field_name = True


class JobSnap(BaseModel):
    """
    Represents a snapshot of a job on a target node.
    """

    id: str = Field(default="")
    kind: str = Field(default="")
    start_time: str = Field(default="", alias="start-time")
    end_time: str = Field(default="", alias="end-time")
    bucket: Optional[BucketModel] = Field(default=None, alias="bck")
    source_bck: Optional[BucketModel] = Field(default=None, alias="src-bck")
    destination_bck: Optional[BucketModel] = Field(default=None, alias="dst-bck")
    stats: Optional[JobStats] = Field(default=None)
    aborted: bool = Field(default=False)
    is_idle: bool = Field(default=False)
    abort_err: str = Field(default="", alias="abort-err")
    packed: int = Field(
        default=0, alias="glob.id"
    )  # Bitwise representation of the number of joggers, workers, and channel full status

    class Config:  # pylint: disable=missing-class-docstring
        allow_population_by_field_name = True

    def unpack_glob_id(self) -> tuple[int, int, int]:
        """
        Unpack the `glob_id` into (njoggers, nworkers, chan_full).
        The packed value is a bitwise representation of the number of joggers, workers, and channel full status.
        """
        njoggers = self.packed & JOGGER_COUNT_MASK
        nworkers = (self.packed & WORKER_COUNT_MASK) >> WORKER_COUNT_SHIFT
        chbuf_cnt = (self.packed & CHANNEL_COUNT_MASK) >> CHANNEL_COUNT_SHIFT
        return njoggers, nworkers, chbuf_cnt


class AggregatedJobSnap(BaseModel):
    """
    Represents job snapshots grouped by target ID.

    Under the hood this is a root model (mapping node IDs →
    lists of JobSnapshot), but we expose it via the `snapshots`
    property so that static checkers don’t get confused.
    """

    __root__: Dict[str, List[JobSnap]] = Field(default_factory=dict)

    @property
    def snapshots(self) -> Dict[str, List[JobSnap]]:
        """
        Return the underlying dict of nodeID → [JobSnapshot,...].
        """
        # __root__ in __dict__ is the real dict, not a FieldInfo
        return self.__dict__.get("__root__", {})

    def list_snapshots(self) -> List[JobSnap]:
        """
        Return a flat list of all JobSnapshot instances.
        """
        return [snap for snaps in self.snapshots.values() for snap in snaps]

    def get_num_workers(self) -> int:
        """
        Return the nworkers from any one snapshot (as all are the same).
        """
        for snaps in self.snapshots.values():
            if snaps:
                # unpack() → (njoggers, nworkers, chan_full)
                return snaps[0].unpack_glob_id()[1]
        return 0


class CopyBckMsg(BaseModel):
    """
    API message structure for copying a bucket
    """

    prefix: str = ""
    prepend: str
    dry_run: bool
    force: bool
    latest: bool
    sync: bool

    def as_dict(self):
        return {
            "prefix": self.prefix,
            "prepend": self.prepend,
            "dry_run": self.dry_run,
            "force": self.force,
            "latest-ver": self.latest,
            "synchronize": self.sync,
        }


class ListObjectsMsg(BaseModel):
    """
    API message structure for listing objects in a bucket
    """

    prefix: str
    page_size: int
    uuid: str
    props: str
    continuation_token: str
    flags: List[ListObjectFlag]
    target: str

    def as_dict(self):
        return {
            "prefix": self.prefix,
            "pagesize": self.page_size,
            "uuid": self.uuid,
            "props": self.props,
            "continuation_token": self.continuation_token,
            "flags": str(ListObjectFlag.join_flags(self.flags)),
            "target": self.target,
        }


class TransformBckMsg(BaseModel):
    """
    API message structure for requesting an etl transform on a bucket
    """

    etl_name: str
    timeout: str

    def as_dict(self):
        return {"id": self.etl_name, "request_timeout": self.timeout}


class TCBckMsg(BaseModel):
    """
    API message structure for transforming or copying between buckets.
    Can be used on its own for an entire bucket or encapsulated in TCMultiObj to apply only to a selection of objects
    """

    ext: Dict[str, str] = None
    copy_msg: CopyBckMsg = None
    transform_msg: TransformBckMsg = None
    num_workers: int = None

    def as_dict(self):
        dict_rep = {}
        if self.ext:
            dict_rep["ext"] = self.ext
        if self.num_workers:
            dict_rep["num-workers"] = self.num_workers
        if self.copy_msg:
            for key, val in self.copy_msg.as_dict().items():
                dict_rep[key] = val
        if self.transform_msg:
            for key, val in self.transform_msg.as_dict().items():
                dict_rep[key] = val
        return dict_rep


class PrefetchMsg(BaseModel):
    """
    API message structure for prefetching objects from remote buckets.
    """

    object_selection: Dict
    continue_on_err: bool
    latest: bool
    blob_threshold: int = None
    num_workers: int = None

    def as_dict(self):
        dict_rep = self.object_selection
        dict_rep["coer"] = self.continue_on_err
        dict_rep["latest-ver"] = self.latest
        if self.blob_threshold:
            dict_rep["blob-threshold"] = self.blob_threshold
        if self.num_workers:
            dict_rep["num-workers"] = self.num_workers
        return dict_rep


class TCMultiObj(BaseModel):
    """
    API message structure for transforming or copying multiple objects between buckets
    """

    to_bck: BucketModel
    tc_msg: TCBckMsg = None
    continue_on_err: bool
    object_selection: Dict
    num_workers: int = None

    def as_dict(self):
        dict_rep = self.object_selection
        if self.tc_msg:
            for key, val in self.tc_msg.as_dict().items():
                dict_rep[key] = val
        dict_rep["tobck"] = self.to_bck.as_dict()
        dict_rep["coer"] = self.continue_on_err
        if self.num_workers:
            dict_rep["num-workers"] = self.num_workers
        return dict_rep


class ArchiveMultiObj(BaseModel):
    """
    API message structure for multi-object archive requests
    """

    archive_name: str
    to_bck: BucketModel
    mime: str = None
    include_source_name = False
    allow_append = False
    continue_on_err = False
    object_selection: Dict

    def as_dict(self):
        dict_rep = self.object_selection
        dict_rep["archname"] = self.archive_name
        dict_rep["isbn"] = self.include_source_name
        dict_rep["aate"] = self.allow_append
        dict_rep["coer"] = self.continue_on_err
        if self.mime:
            dict_rep["mime"] = self.mime
        if self.to_bck:
            dict_rep["tobck"] = self.to_bck.as_dict()
        return dict_rep


class BlobMsg(BaseModel):
    """
    Represents the set of args the sdk will pass to AIStore when making a blob-download request
    and provides conversion to the expected json format
    """

    chunk_size: int = None
    num_workers: int = None
    latest: bool

    def as_dict(self):
        dict_rep = {
            "latest-ver": self.latest,
        }
        if self.chunk_size:
            dict_rep["chunk-size"] = self.chunk_size
        if self.num_workers:
            dict_rep["num-workers"] = self.num_workers
        return dict_rep



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
from dataclasses import dataclass
from typing import Dict, Optional, Union


@dataclass
class ETLConfig:
    """
    Configuration for ETL operations.

    Attributes:
        name (str): Name of the ETL pipeline.
        args (Optional[Union[str, Dict]]): Optional parameters for configuring the ETL pipeline.
            Can be provided as:
                - A string for simple arguments.
                - A dictionary for structured key-value configurations.
            Defaults to an empty dictionary.

    Example:
        etl_config = ETLConfig(
            name="image-transform",
            args={"format": "jpeg", "resize": "256x256"}
        )
    """

    name: str
    args: Optional[Union[str, Dict]] = None



#
# Copyright (c) 2018-2024, NVIDIA CORPORATION. All rights reserved.
#
from typing import Optional, Tuple, Union
import os
import warnings
from urllib3 import Retry

from aistore.sdk.bucket import Bucket
from aistore.sdk.provider import Provider
from aistore.sdk.const import AIS_AUTHN_TOKEN
from aistore.sdk.cluster import Cluster
from aistore.sdk.dsort import Dsort
from aistore.sdk.request_client import RequestClient
from aistore.sdk.session_manager import SessionManager
from aistore.sdk.types import Namespace
from aistore.sdk.job import Job
from aistore.sdk.etl.etl import Etl
from aistore.sdk.utils import parse_url
from aistore.sdk.obj.object import Object
from aistore.sdk.errors import InvalidURLException
from aistore.sdk.retry_config import RetryConfig


class Client:
    """
    AIStore client for managing buckets, objects, and ETL jobs.

    Args:
        endpoint (str): AIStore endpoint.
        skip_verify (bool, optional): If True, skip SSL certificate verification. Defaults to False.
        ca_cert (str, optional): Path to a CA certificate file for SSL verification. If not provided,
            the 'AIS_CLIENT_CA' environment variable will be used. Defaults to None.
        client_cert (Union[str, Tuple[str, str], None], optional): Path to a client certificate PEM file
            or a tuple (cert, key) for mTLS. If not provided, 'AIS_CRT' and 'AIS_CRT_KEY' environment
            variables will be used. Defaults to None.
        timeout (Union[float, Tuple[float, float], None], optional): Timeout for HTTP requests.
            - Single float (e.g., `5.0`): Applies to both connection and read timeouts.
            - Tuple (e.g., `(3.0, 20.0)`): First value is the connection timeout, second is the read timeout.
            - `None`: Disables timeouts (not recommended). Defaults to `(3, 20)`.
        retry_config (RetryConfig, optional): Defines retry behavior for HTTP and network failures.
            If not provided, the default retry configuration (`RetryConfig.default()`) is used.
        retry (urllib3.Retry, optional): [Deprecated] Retry configuration from urllib3. Use `retry_config` instead.
        token (str, optional): Authorization token. If not provided, the 'AIS_AUTHN_TOKEN' environment variable
            will be used. Defaults to None.
        max_pool_size (int, optional): Maximum number of connections per host in the connection pool.
            Defaults to 10.
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(
        self,
        endpoint: str,
        skip_verify: bool = False,
        ca_cert: Optional[str] = None,
        client_cert: Optional[Union[str, Tuple[str, str]]] = None,
        timeout: Optional[Union[float, Tuple[float, float]]] = (3, 20),
        retry_config: Optional[RetryConfig] = None,
        retry: Optional[Retry] = None,  # deprecated
        token: Optional[str] = None,
        max_pool_size: int = 10,
    ):
        # Use `retry_config` (RetryConfig) if provided; otherwise, fall back to the deprecated `retry` (urllib3.Retry)
        if retry_config is not None:
            self.retry_config = retry_config
        elif retry is not None:
            warnings.warn(
                "'retry' is deprecated and will be removed in a future release. "
                "Use 'retry_config' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.retry_config = RetryConfig.default()
            self.retry_config.http_retry = retry
        else:
            self.retry_config = RetryConfig.default()

        session_manager = SessionManager(
            retry=self.retry_config.http_retry,
            ca_cert=ca_cert,
            client_cert=client_cert,
            skip_verify=skip_verify,
            max_pool_size=max_pool_size,
        )

        # Check for token from arguments or environment variable
        if not token:
            token = os.environ.get(AIS_AUTHN_TOKEN)

        self._request_client = RequestClient(
            endpoint=endpoint,
            session_manager=session_manager,
            timeout=timeout,
            token=token,
            network_retry_config=self.retry_config.network_retry,
        )

    def bucket(
        self,
        bck_name: str,
        provider: Union[Provider, str] = Provider.AIS,
        namespace: Namespace = None,
    ):
        """
        Factory constructor for bucket object.
        Does not make any HTTP request, only instantiates a bucket object.

        Args:
            bck_name (str): Name of bucket
            provider (str or Provider): Provider of bucket, one of "ais", "aws", "gcp", ...
                (optional, defaults to ais)
            namespace (Namespace): Namespace of bucket (optional, defaults to None)

        Returns:
            The bucket object created.
        """
        return Bucket(
            client=self._request_client,
            name=bck_name,
            provider=provider,
            namespace=namespace,
        )

    def cluster(self):
        """
        Factory constructor for cluster object.
        Does not make any HTTP request, only instantiates a cluster object.

        Returns:
            The cluster object created.
        """
        return Cluster(client=self._request_client)

    def job(self, job_id: str = "", job_kind: str = ""):
        """
        Factory constructor for job object, which contains job-related functions.
        Does not make any HTTP request, only instantiates a job object.

        Args:
            job_id (str, optional): Optional ID for interacting with a specific job
            job_kind (str, optional): Optional specific type of job empty for all kinds

        Returns:
            The job object created.
        """
        return Job(client=self._request_client, job_id=job_id, job_kind=job_kind)

    def etl(self, etl_name: str):
        """
        Factory constructor for ETL object.
        Contains APIs related to AIStore ETL operations.
        Does not make any HTTP request, only instantiates an ETL object.

        Args:
            etl_name (str): Name of the ETL

        Returns:
            The ETL object created.
        """
        return Etl(client=self._request_client, name=etl_name)

    def dsort(self, dsort_id: str = ""):
        """
        Factory constructor for dSort object.
        Contains APIs related to AIStore dSort operations.
        Does not make any HTTP request, only instantiates a dSort object.

        Args:
            dsort_id: ID of the dSort job

        Returns:
            dSort object created
        """
        return Dsort(client=self._request_client, dsort_id=dsort_id)

    def fetch_object_by_url(self, url: str) -> Object:
        """
        Deprecated: Use `get_object_from_url` instead.

        Creates an Object instance from a URL.

        This method does not make any HTTP requests.

        Args:
            url (str): Full URL of the object (e.g., "ais://bucket1/file.txt")

        Returns:
            Object: The object constructed from the specified URL
        """
        warnings.warn(
            "The 'fetch_object_by_url' method is deprecated and will be removed in a future release. "
            "Please use 'get_object_from_url' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_object_from_url(url)

    def get_object_from_url(self, url: str) -> Object:
        """
        Creates an Object instance from a URL.

        This method does not make any HTTP requests.

        Args:
            url (str): Full URL of the object (e.g., "ais://bucket1/file.txt")

        Returns:
            Object: The object constructed from the specified URL

        Raises:
            InvalidURLException: If the URL is invalid.
        """
        try:
            provider, bck_name, obj_name = parse_url(url)
            if not provider or not bck_name or not obj_name:
                raise InvalidURLException(url)
            return self.bucket(bck_name, provider=provider).object(obj_name)
        except InvalidURLException as err:
            raise err



#
# Copyright (c) 2024-2025, NVIDIA CORPORATION. All rights reserved.
#

from typing import Optional, Tuple, Union
from urllib3 import Retry

from aistore.sdk.authn.response_handler import AuthNResponseHandler
from aistore.sdk.request_client import RequestClient
from aistore.sdk.session_manager import SessionManager
from aistore.sdk.utils import get_logger
from aistore.sdk.authn.types import TokenMsg, LoginMsg
from aistore.sdk.authn.cluster_manager import ClusterManager
from aistore.sdk.authn.role_manager import RoleManager
from aistore.sdk.authn.token_manager import TokenManager
from aistore.sdk.authn.user_manager import UserManager
from aistore.sdk.const import (
    HTTP_METHOD_POST,
    URL_PATH_AUTHN_USERS,
)

logger = get_logger(__name__)


class AuthNClient:
    """
    AuthN client for managing authentication.

    This client provides methods to interact with AuthN Server.
    For more info on AuthN Server, see https://github.com/NVIDIA/aistore/blob/main/docs/authn.md

    Args:
        endpoint (str): AuthN service endpoint URL.
        skip_verify (bool, optional): If True, skip SSL certificate verification. Defaults to False.
        ca_cert (str, optional): Path to a CA certificate file for SSL verification.
        timeout (Union[float, Tuple[float, float], None], optional): Request timeout in seconds; a single float
            for both connect/read timeouts (e.g., 5.0), a tuple for separate connect/read timeouts (e.g., (3.0, 10.0)),
            or None to disable timeout.
        retry (urllib3.Retry, optional): Retry configuration object from the urllib3 library.
        token (str, optional): Authorization token.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        endpoint: str,
        skip_verify: bool = False,
        ca_cert: Optional[str] = None,
        timeout: Optional[Union[float, Tuple[float, float]]] = None,
        retry: Optional[Retry] = None,
        token: Optional[str] = None,
    ):
        logger.info("Initializing AuthNClient")
        session_manager = SessionManager(
            retry=retry, ca_cert=ca_cert, skip_verify=skip_verify
        )
        self._request_client = RequestClient(
            endpoint=endpoint,
            session_manager=session_manager,
            timeout=timeout,
            token=token,
            response_handler=AuthNResponseHandler(),
        )
        logger.info("AuthNClient initialized with endpoint: %s", endpoint)

    @property
    def client(self) -> RequestClient:
        """
        Get the request client.

        Returns:
            RequestClient: The client this AuthN client uses to make requests.
        """
        return self._request_client

    def login(
        self,
        username: str,
        password: str,
        expires_in: Optional[Union[int, float]] = None,
    ) -> str:
        """
        Logs in to the AuthN Server and returns an authorization token.

        Args:
            username (str): The username to log in with.
            password (str): The password to log in with.
            expires_in (Optional[Union[int, float]]): The expiration duration of the token in seconds.

        Returns:
            str: An authorization token to use for future requests.

        Raises:
            ValueError: If the password is empty or consists only of spaces.
            AISError: If the login request fails.
        """
        if password.strip() == "":
            raise ValueError("Password cannot be empty or spaces only")

        logger.info("Attempting to log in with username: %s", username)
        login_msg = LoginMsg(password=password, expires_in=expires_in).as_dict()

        try:
            token = self.client.request_deserialize(
                HTTP_METHOD_POST,
                path=f"{URL_PATH_AUTHN_USERS}/{username}",
                json=login_msg,
                res_model=TokenMsg,
            ).token
            logger.info("Login successful for username: %s", username)
            # Update the client token
            self.client.token = token
            return token
        except Exception as err:
            logger.error("Login failed for username: %s, error: %s", username, err)
            raise

    def logout(self) -> None:
        """
        Logs out and revokes current token from the AuthN Server.

        Raises:
            AISError: If the logout request fails.
        """
        if not self.client.token:
            raise ValueError("Must be logged in first (no token)")

        try:
            logger.info("Logging out")
            self.token_manager().revoke(token=self.client.token)
            self.client.token = None
        except Exception as err:
            logger.error("Logout failed, error: %s", err)
            raise

    def cluster_manager(self) -> ClusterManager:
        """
        Factory method to create a ClusterManager instance.

        Returns:
            ClusterManager: An instance to manage cluster operations.
        """
        return ClusterManager(client=self._request_client)

    def role_manager(self) -> RoleManager:
        """
        Factory method to create a RoleManager instance.

        Returns:
            RoleManager: An instance to manage role operations.
        """
        return RoleManager(client=self._request_client)

    def user_manager(self) -> UserManager:
        """
        Factory method to create a UserManager instance.

        Returns:
            UserManager: An instance to manage user operations.
        """
        return UserManager(client=self._request_client)

    def token_manager(self) -> TokenManager:
        """
        Factory method to create a TokenManager instance.

        Returns:
            TokenManager: An instance to manage token operations.
        """
        return TokenManager(client=self._request_client)



from typing import Literal, Union, Optional
from pathlib import Path

import json
import yaml

from pydantic.v1 import BaseModel, root_validator

from aistore.sdk.types import BucketModel
from aistore.sdk.multiobj import ObjectNames, ObjectRange
from aistore.sdk.dsort.ekm import ExternalKeyMap


# pylint: disable=redefined-builtin,too-few-public-methods
class DsortShardsGroup:
    """
    Represents the configuration for the input or output of a shard group in a dSort job
    """

    def __init__(
        self,
        bck: BucketModel,
        role: Literal["input", "output"],
        format: Union[ObjectNames, ObjectRange, ExternalKeyMap],
        extension: str,
    ):
        if role not in ["input", "output"]:
            raise ValueError('role must be either "input" or "output"')
        if role == "input" and not isinstance(format, (ObjectNames, ObjectRange)):
            raise ValueError(
                'format of input role must be an instance of "ObjectNames" or "ObjectRange"'
            )
        if role == "output" and not isinstance(format, (ObjectRange, ExternalKeyMap)):
            raise ValueError(
                'format of output role must be an instance of "ObjectRange" or "ExternalKeyMap"'
            )

        self.bck = bck
        self.role = role
        self.extension = extension
        self.format = format

    def as_dict(self):
        """
        Converts the DsortShardsGroup instance to a dictionary representation
        """

        return {
            f"{self.role}_bck": self.bck.as_dict(),
            f"{self.role}_format": (
                self.format.get_value() if self.role == "input" else str(self.format)
            ),
            f"{self.role}_extension": self.extension,
        }


class DsortAlgorithm(BaseModel):
    """
    Represents the algorithm used in a Dsort job
    """

    kind: Literal["alphanumeric", "shuffle", "content"] = "alphanumeric"
    decreasing: bool = False
    seed: Optional[str] = ""
    extension: Optional[str] = None
    content_key_type: Optional[Literal["int", "float", "string"]] = None

    # pylint: disable=no-self-argument
    @root_validator
    def validate_content_fields(cls, values):
        """
        Validates required key fields
        """
        kind = values.get("kind")
        extension = values.get("extension")
        content_key_type = values.get("content_key_type")

        if kind == "content":
            if not extension:
                raise ValueError(
                    'For kind="content", the "extension" field is required.'
                )
            if not content_key_type:
                raise ValueError(
                    'For kind="content", the "content_key_type" field is required.'
                )
        else:
            if extension or content_key_type:
                raise ValueError(
                    'The "extension" and "content_key_type" fields are only allowed for kind="content".'
                )

        return values

    def as_dict(self):
        """
        Converts the DsortAlgorithm instance to a dictionary representation
        """
        dict_rep = {"kind": self.kind, "decreasing": self.decreasing, "seed": self.seed}
        if self.kind == "content":
            dict_rep["extension"] = self.extension
            dict_rep["content_key_type"] = self.content_key_type
        return dict_rep


class DsortFramework:
    """
    Represents the framework for a dSort job, including input and output shard configurations.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        input_shards: DsortShardsGroup,
        output_shards: DsortShardsGroup,
        algorithm: DsortAlgorithm = None,
        description=None,
        output_shard_size=None,
    ) -> None:
        self.input_shards = input_shards
        self.output_shards = output_shards
        self.output_shard_size = output_shard_size
        if algorithm is None:
            algorithm = DsortAlgorithm()
        self.algorithm = algorithm
        self.description = description

        self.ekm_file = None
        self.ekm_sep = None

    @classmethod
    def from_file(cls, spec):
        """
        Class method to create a DsortFramework instance from a JSON or YAML file

        Args:
            spec (str or Path): The path to the JSON or YAML file containing the specification

        Returns:
            DsortFramework: An instance of the DsortFramework class

        Raises:
            ValueError: If the file extension is not .json, .yml, or .yaml
        """
        spec_path = Path(spec)
        ext = spec_path.suffix.lower()
        if ext not in [".json", ".yaml", ".yml"]:
            raise ValueError(
                "Unsupported file format. Please provide a .json, .yml, or .yaml file."
            )
        load_func = json.load if ext == ".json" else yaml.safe_load
        with open(spec, "r", encoding="utf-8") as file_data:
            spec_data = load_func(file_data)

            # Parse and validate input_format
            input_format = spec_data.get("input_format", None)
            if not isinstance(input_format, dict):
                raise ValueError("input_format must be a dictionary")

            if "template" in input_format:
                input_format = ObjectRange.from_string(input_format["template"])
            elif "objnames" in input_format:
                input_format = ObjectNames(input_format["objnames"])
            else:
                raise ValueError(
                    '"input_format" dictionary must contain either the key "template" or "objnames"'
                )

            # Parse and validate output_format
            output_format = spec_data.get("output_format", None)
            if isinstance(output_format, list):
                output_format = ObjectNames(output_format)
            elif isinstance(output_format, str):
                output_format = ObjectRange.from_string(output_format)
            else:
                raise ValueError(
                    '"input_format" string must be either a list or a range string'
                )

            framework = cls(
                input_shards=DsortShardsGroup(
                    bck=BucketModel(**spec_data.get("input_bck")),
                    role="input",
                    format=input_format,
                    extension=spec_data.get("input_extension", ""),
                ),
                output_shards=DsortShardsGroup(
                    bck=BucketModel(**spec_data.get("output_bck")),
                    role="output",
                    format=output_format,
                    extension=spec_data.get("output_extension", ""),
                ),
                algorithm=DsortAlgorithm(**spec_data.get("algorithm", {})),
                output_shard_size=spec_data.get("output_shard_size", ""),
            )

            framework.ekm_file = spec_data.get("ekm_file", "")
            framework.ekm_sep = spec_data.get("ekm_sep", "")

            return framework

    def to_spec(self):
        """
        Converts the DsortFramework instance to a dictionary representation for use in dSort job specification

        Returns:
            Dictionary representation of dSort specification
        """
        spec = {**self.input_shards.as_dict(), **self.output_shards.as_dict()}
        spec["algorithm"] = self.algorithm.as_dict()
        spec["description"] = self.description if self.description else ""
        if self.output_shard_size:
            spec["output_shard_size"] = self.output_shard_size

        if self.ekm_file:
            spec["ekm_file"] = self.ekm_file
            spec["ekm_sep"] = self.ekm_sep

        return spec



#
# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
#

from dataclasses import dataclass
from aistore.sdk.client import Client
from aistore.sdk.provider import Provider


@dataclass
class DataShard:
    """
    A class representing a group of data shards stored in AIStore

    Attributes:
        client_url (str): AIS endpoint URL
        bucket_name (str): The name of the bucket containing the data shards
        provider (str): The provider of the bucket, default is "ais"
        prefix (str): The prefix for the data shards, default is ""
        etl_name (str): The name of the ETL to apply to the data shards, default is None
    """

    client_url: str
    bucket_name: str
    provider: str = Provider.AIS
    prefix: str = ""
    etl_name: str = None

    def __post_init__(self):
        self.client = Client(self.client_url)
        self.bucket = self.client.bucket(
            bck_name=self.bucket_name, provider=self.provider
        )



#
# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
#

from typing import Iterator, Optional, Tuple

from aistore.sdk.obj.content_iterator import ContentIterator
from aistore.sdk.utils import get_logger
from aistore.sdk.obj.obj_file.errors import (
    ObjectFileReaderStreamError,
    ObjectFileReaderMaxResumeError,
)

logger = get_logger(__name__)


def reset_iterator(
    content_iterator: ContentIterator,
    resume_position: Optional[int] = 0,
) -> Iterator[bytes]:
    """
    Return a new iterator for establishing an object stream and reading chunks of data from
    byte position `resume_position`.

    Args:
        content_iterator (ContentIterator): An instance of `ContentIterator` to read data from.
        resume_position (int, optional): The byte position to resume reading from. Defaults to 0.

    Returns:
        Iterator[bytes]: An iterator to read chunks of data from the object stream.

    Raises:
        ObjectFileReaderStreamError if a connection cannot be made.
    """
    try:
        return content_iterator.iter(offset=resume_position)
    except Exception as err:
        logger.error("Error establishing object stream: (%s)", err)
        raise ObjectFileReaderStreamError(err) from err


def increment_resume(resume_total: int, max_resume: int, err: Exception) -> int:
    """
    Increment the number of resume attempts and raise an error if the maximum allowed is exceeded.

    Args:
        resume_total (int): The number of resume attempts made.
        max_resume (int): The maximum number of resume attempts allowed.
        err (Exception): The error that triggered the resume attempt.

    Returns:
        int: The updated number of resume attempts.

    Raises:
        ObjectFileReaderMaxResumeError: If the number of resume attempts exceeds the maximum allowed.
    """
    resume_total += 1
    if resume_total > max_resume:
        raise ObjectFileReaderMaxResumeError(err, resume_total) from err
    return resume_total


def handle_chunked_encoding_error(
    content_iterator: ContentIterator,
    resume_position: int,
    resume_total: int,
    max_resume: int,
    err: Exception,
) -> Tuple[Iterator[bytes], int]:
    """
    Handle the chunked encoding error by incrementing the resume count, logging a warning,
    and resetting the iterator from the last known position.

    Args:
        content_iterator (ContentIterator): The content iterator used to read the data.
        resume_position (int): The byte position from which to resume reading.
        resume_total (int): The current number of resume attempts.
        max_resume (int): The maximum number of resume attempts allowed.
        err (Exception): The error that caused the resume attempt.

    Returns:
        Tuple[Iterator[bytes], int]: The new iterator and the updated resume total.

    Raises:
        ObjectFileReaderMaxResumeError: If the maximum number of resume attempts is exceeded.
    """
    resume_total = increment_resume(resume_total, max_resume, err)
    logger.warning(
        "Chunked encoding error (%s), retrying %d/%d",
        err,
        resume_total,
        max_resume,
    )
    chunk_iterator = reset_iterator(content_iterator, resume_position)
    return chunk_iterator, resume_total



#
# Copyright (c) 2022-2025, NVIDIA CORPORATION. All rights reserved.
#

import warnings

from dataclasses import dataclass
from io import BufferedWriter
from pathlib import Path
from typing import Dict, Optional
import os
from json import dumps as json_dumps
from urllib.parse import quote

from requests import Response
from requests.structures import CaseInsensitiveDict
from aistore.sdk.archive_config import ArchiveConfig
from aistore.sdk.blob_download_config import BlobDownloadConfig
from aistore.sdk.etl import ETLConfig
from aistore.sdk.const import (
    BYTE_RANGE_PREFIX_LENGTH,
    DEFAULT_CHUNK_SIZE,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_HEAD,
    QPARAM_ARCHPATH,
    QPARAM_ARCHREGX,
    QPARAM_ARCHMODE,
    QPARAM_ETL_NAME,
    QPARAM_ETL_ARGS,
    QPARAM_LATEST,
    ACT_PROMOTE,
    HTTP_METHOD_POST,
    URL_PATH_OBJECTS,
    HEADER_RANGE,
    ACT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_WORKERS,
    HEADER_OBJECT_BLOB_CHUNK_SIZE,
)
from aistore.sdk.provider import Provider
from aistore.sdk.obj.object_client import ObjectClient
from aistore.sdk.obj.object_reader import ObjectReader
from aistore.sdk.obj.object_writer import ObjectWriter
from aistore.sdk.request_client import RequestClient
from aistore.sdk.types import (
    ActionMsg,
    PromoteAPIArgs,
    BlobMsg,
)
from aistore.sdk.obj.object_props import ObjectProps


@dataclass
class BucketDetails:
    """
    Metadata about a bucket, used by objects within that bucket.
    """

    name: str
    provider: Provider
    qparams: Dict[str, str]
    path: str


class Object:
    """
    Provides methods for interacting with an object in AIS.

    Args:
        client (RequestClient): Client used for all http requests.
        bck_details (BucketDetails): Metadata about the bucket to which this object belongs.
        name (str): Name of the object.
        props (ObjectProps, optional): Properties of the object, as updated by head(), optionally pre-initialized.
    """

    def __init__(
        self,
        client: RequestClient,
        bck_details: BucketDetails,
        name: str,
        props: ObjectProps = None,
    ):
        self._client = client
        self._bck_details = bck_details
        self._bck_path = f"{URL_PATH_OBJECTS}/{ bck_details.name}"
        self._name = name
        self._props = props
        self._object_path = f"{self._bck_path}/{quote(name)}"

    @property
    def bucket_name(self) -> str:
        """Name of the bucket where this object resides."""
        return self._bck_details.name

    @property
    def bucket_provider(self) -> Provider:
        """Provider of the bucket where this object resides (e.g. ais, s3, gcp)."""
        return self._bck_details.provider

    @property
    def query_params(self) -> Dict[str, str]:
        """Query params used as a base for constructing all requests for this object."""
        return self._bck_details.qparams

    @property
    def name(self) -> str:
        """Name of this object."""
        return self._name

    @property
    def props(self) -> ObjectProps:
        """
        Get the latest properties of the object.

        This will make a HEAD request to the AIStore cluster to fetch up-to-date object headers
        and refresh the internal `_props` cache. Use this when you want to ensure you're accessing
        the most recent metadata for the object.

        Returns:
            ObjectProps: The latest object properties from the server.
        """
        self.head()
        return self._props

    @property
    def props_cached(self) -> Optional[ObjectProps]:
        """
        Get the cached object properties (without making a network call).

        This is useful when:
        - You want to avoid a network request.
        - You're sure the cached `_props` was already set via a previous call to `head()` or during object construction.

        Returns:
            ObjectProps or None: Cached object properties, or None if not set.
        """
        return self._props

    def head(self) -> CaseInsensitiveDict:
        """
        Requests object properties and returns headers. Updates props.

        Returns:
            Response header with the object properties.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        headers = self._client.request(
            HTTP_METHOD_HEAD,
            path=self._object_path,
            params=self.query_params,
        ).headers
        self._props = ObjectProps(headers)
        return headers

    # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals
    def get_reader(
        self,
        archive_config: Optional[ArchiveConfig] = None,
        blob_download_config: Optional[BlobDownloadConfig] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl: Optional[ETLConfig] = None,
        writer: Optional[BufferedWriter] = None,
        latest: bool = False,
        byte_range: Optional[str] = None,
        direct: bool = False,
    ) -> ObjectReader:
        """
        Creates and returns an ObjectReader with access to object contents
        and optionally writes to a provided writer.

        Args:
            archive_config (Optional[ArchiveConfig]): Settings for archive extraction.
            blob_download_config (Optional[BlobDownloadConfig]): Settings for using blob download.
            chunk_size (int, optional): Chunk size to use while reading from stream.
            etl (Optional[ETLConfig]): Settings for ETL-specific operations (name, args).
            writer (Optional[BufferedWriter]): User-provided writer for writing content output.
                The user is responsible for closing the writer.
            latest (bool, optional): GET the latest object version from the associated remote bucket.
            byte_range (Optional[str]): Byte range in RFC 7233 format for single-range requests
                (e.g., "bytes=0-499", "bytes=500-", "bytes=-500").
                See: https://www.rfc-editor.org/rfc/rfc7233#section-2.1.
            direct (bool, optional): If True, the object content is read directly from the target node,
                bypassing the proxy.

        Returns:
            ObjectReader: An iterator for streaming object content.

        Raises:
            ValueError: If Byte Range is used with Blob Download.
            requests.RequestException: If an error occurs during the request.
            requests.ConnectionError: If there is a connection error.
            requests.ConnectionTimeout: If the connection times out.
            requests.ReadTimeout: If the read operation times out.
        """

        params = self.query_params.copy()
        headers = {}
        byte_range_tuple = (None, None)

        # Archive Configuration
        if archive_config:
            if archive_config.mode:
                params[QPARAM_ARCHMODE] = archive_config.mode.value
            params.update(
                {
                    QPARAM_ARCHPATH: archive_config.archpath,
                    QPARAM_ARCHREGX: archive_config.regex,
                }
            )

        # Blob Download Configuration
        if blob_download_config:
            headers.update(
                {
                    HEADER_OBJECT_BLOB_DOWNLOAD: "true",
                    HEADER_OBJECT_BLOB_CHUNK_SIZE: blob_download_config.chunk_size,
                    HEADER_OBJECT_BLOB_WORKERS: blob_download_config.num_workers,
                }
            )

        # ETL Configuration
        if etl:
            params[QPARAM_ETL_NAME] = etl.name
            params[QPARAM_ETL_ARGS] = (
                json_dumps(etl.args, separators=(",", ":"))
                if isinstance(etl.args, dict)
                else etl.args
            )

        # Latest Object Version
        if latest:
            params[QPARAM_LATEST] = "true"

        # Byte Range Validation
        if byte_range and blob_download_config:
            raise ValueError("Cannot use Byte Range with Blob Download.")

        if byte_range:
            # For range formatting, see the spec:
            # https://www.rfc-editor.org/rfc/rfc7233#section-2.1
            headers = {HEADER_RANGE: byte_range}
            # Extract left (range_l) and right (range_r) bounds from the byte range string
            headers[HEADER_RANGE] = byte_range
            byte_range_l, _, byte_range_r = byte_range[
                BYTE_RANGE_PREFIX_LENGTH:
            ].partition("-")
            byte_range_tuple = (
                int(byte_range_l) if byte_range_l else None,
                int(byte_range_r) if byte_range_r else None,
            )

        # Object Client
        obj_client = ObjectClient(
            request_client=self._client,
            path=self._object_path,
            params=params,
            headers=headers,
            byte_range=byte_range_tuple,
            uname=os.path.join(self._bck_details.path, self.name) if direct else None,
        )

        obj_reader = ObjectReader(object_client=obj_client, chunk_size=chunk_size)

        if writer:
            writer.writelines(obj_reader)

        return obj_reader

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def get(
        self,
        archive_config: ArchiveConfig = None,
        blob_download_config: BlobDownloadConfig = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl: ETLConfig = None,
        writer: BufferedWriter = None,
        latest: bool = False,
        byte_range: str = None,
    ) -> ObjectReader:
        """
        Deprecated: Use 'get_reader' instead.

        Creates and returns an ObjectReader with access to object contents and optionally writes to a provided writer.

        Args:
            archive_config (ArchiveConfig, optional): Settings for archive extraction.
            blob_download_config (BlobDownloadConfig, optional): Settings for using blob download.
            chunk_size (int, optional): Chunk size to use while reading from stream.
            etl (ETLConfig, optional): Settings for ETL-specific operations (name, meta).
            writer (BufferedWriter, optional): User-provided writer for writing content output.
                The user is responsible for closing the writer.
            latest (bool, optional): GET the latest object version from the associated remote bucket.
            byte_range (str, optional): Byte range in RFC 7233 format for single-range requests
                (e.g., "bytes=0-499", "bytes=500-", "bytes=-500").
                See: https://www.rfc-editor.org/rfc/rfc7233#section-2.1.

        Returns:
            ObjectReader: An ObjectReader that can be iterated over to stream chunks of object content
            or used to read all content directly.

        Raises:
            ValueError: If Byte Range is used with Blob Download.
            requests.RequestException: If an error occurs during the request.
            requests.ConnectionError: If there is a connection error.
            requests.ConnectionTimeout: If the connection times out.
            requests.ReadTimeout: If the read operation times out.
        """
        warnings.warn(
            "The 'get' method is deprecated and will be removed in a future release. "
            "Please use 'get_reader' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_reader(
            archive_config=archive_config,
            blob_download_config=blob_download_config,
            chunk_size=chunk_size,
            etl=etl,
            writer=writer,
            latest=latest,
            byte_range=byte_range,
        )

    def get_semantic_url(self) -> str:
        """
        Get the semantic URL to the object

        Returns:
            Semantic URL to get object
        """

        return f"{self.bucket_provider.value}://{self.bucket_name}/{self._name}"

    def get_url(self, archpath: str = "", etl: ETLConfig = None) -> str:
        """
        Get the full url to the object including base url and any query parameters

        Args:
            archpath (str, optional): If the object is an archive, use `archpath` to extract a single file
                from the archive
            etl (ETLConfig, optional): Settings for ETL-specific operations (name, meta).

        Returns:
            Full URL to get object

        """
        params = self.query_params.copy()
        if archpath:
            params[QPARAM_ARCHPATH] = archpath

        # ETL Configuration
        if etl:
            params[QPARAM_ETL_NAME] = etl.name
            if etl.args:
                params[QPARAM_ETL_ARGS] = etl.args

        return self._client.get_full_url(self._object_path, params)

    def put_content(self, content: bytes) -> Response:
        """
        Deprecated: Use 'ObjectWriter.put_content' instead.

        Puts bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to put as an object.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        warnings.warn(
            "The 'put_content' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.put_content' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().put_content(content)

    def put_file(self, path: str or Path) -> Response:
        """
        Deprecated: Use 'ObjectWriter.put_file' instead.

        Puts a local file as an object to a bucket in AIS storage.

        Args:
            path (str or Path): Path to local file

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            ValueError: The path provided is not a valid file
        """
        warnings.warn(
            "The 'put_file' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.put_file' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().put_file(path)

    def get_writer(self) -> ObjectWriter:
        """
        Create an ObjectWriter to write to object contents and attributes.

        Returns:
            An ObjectWriter which can be used to write to an object's contents and attributes.
        """
        return ObjectWriter(self._client, self._object_path, self.query_params)

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def promote(
        self,
        path: str,
        target_id: str = "",
        recursive: bool = False,
        overwrite_dest: bool = False,
        delete_source: bool = False,
        src_not_file_share: bool = False,
    ) -> str:
        """
        Promotes a file or folder an AIS target can access to a bucket in AIS storage.
        These files can be either on the physical disk of an AIS target itself or on a network file system
        the cluster can access.
        See more info here: https://aiatscale.org/blog/2022/03/17/promote

        Args:
            path (str): Path to file or folder the AIS cluster can reach
            target_id (str, optional): Promote files from a specific target node
            recursive (bool, optional): Recursively promote objects from files in directories inside the path
            overwrite_dest (bool, optional): Overwrite objects already on AIS
            delete_source (bool, optional): Delete the source files when done promoting
            src_not_file_share (bool, optional): Optimize if the source is guaranteed to not be on a file share

        Returns:
            Job ID (as str) that can be used to check the status of the operation, or empty if job is done synchronously

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            AISError: Path does not exist on the AIS cluster storage
        """
        value = PromoteAPIArgs(
            source_path=path,
            object_name=self.name,
            target_id=target_id,
            recursive=recursive,
            overwrite_dest=overwrite_dest,
            delete_source=delete_source,
            src_not_file_share=src_not_file_share,
        ).as_dict()
        json_val = ActionMsg(action=ACT_PROMOTE, name=path, value=value).dict()

        return self._client.request(
            HTTP_METHOD_POST,
            path=self._bck_path,
            params=self.query_params,
            json=json_val,
        ).text

    def delete(self) -> Response:
        """
        Delete an object from a bucket.

        Returns:
            None

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        return self._client.request(
            HTTP_METHOD_DELETE,
            path=self._object_path,
            params=self.query_params,
        )

    def blob_download(
        self,
        chunk_size: int = None,
        num_workers: int = None,
        latest: bool = False,
    ) -> str:
        """
        A special facility to download very large remote objects a.k.a. BLOBs
        Returns job ID that for the blob download operation.

        Args:
            chunk_size (int): chunk size in bytes
            num_workers (int): number of concurrent blob-downloading workers (readers)
            latest (bool): GET the latest object version from the associated remote bucket

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
        """
        params = self.query_params.copy()
        value = BlobMsg(
            chunk_size=chunk_size,
            num_workers=num_workers,
            latest=latest,
        ).as_dict()
        json_val = ActionMsg(
            action=ACT_BLOB_DOWNLOAD, value=value, name=self.name
        ).dict()
        return self._client.request(
            HTTP_METHOD_POST, path=self._bck_path, params=params, json=json_val
        ).text

    def append_content(
        self, content: bytes, handle: str = "", flush: bool = False
    ) -> str:
        """
        Deprecated: Use 'ObjectWriter.append_content' instead.

        Append bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to append to the object.
            handle (str): Handle string to use for subsequent appends or flush (empty for the first append).
            flush (bool): Whether to flush and finalize the append operation, making the object accessible.

        Returns:
            handle (str): Handle string to pass for subsequent appends or flush.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        warnings.warn(
            "The 'append_content' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.append_content' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().append_content(content, handle, flush)

    def set_custom_props(
        self, custom_metadata: Dict[str, str], replace_existing: bool = False
    ) -> Response:
        """
        Deprecated: Use 'ObjectWriter.set_custom_props' instead.

        Set custom properties for the object.

        Args:
            custom_metadata (Dict[str, str]): Custom metadata key-value pairs.
            replace_existing (bool, optional): Whether to replace existing metadata. Defaults to False.
        """
        warnings.warn(
            "The 'set_custom_props' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.set_custom_props' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().set_custom_props(custom_metadata, replace_existing)



from aistore.pytorch.map_dataset import AISMapDataset
from aistore.pytorch.multishard_dataset import AISMultiShardStream
from aistore.pytorch.iter_dataset import AISIterDataset
from aistore.pytorch.shard_reader import AISShardReader
from aistore.pytorch.base_map_dataset import AISBaseMapDataset
from aistore.pytorch.base_iter_dataset import AISBaseIterDataset
from aistore.pytorch.dynamic_sampler import DynamicBatchSampler



#
# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
from typing import Type

import requests

from aistore.sdk.authn.errors import (
    AuthNError,
    ErrUserInvalidCredentials,
    ErrClusterNotFound,
    ErrRoleNotFound,
    ErrUserNotFound,
    ErrClusterAlreadyRegistered,
    ErrRoleAlreadyExists,
    ErrUserAlreadyExists,
)
from aistore.sdk.response_handler import ResponseHandler


class AuthNResponseHandler(ResponseHandler):
    """
    Handle responses from an AuthN server
    """

    @property
    def exc_class(self) -> Type[AuthNError]:
        return AuthNError

    def parse_error(self, r: requests.Response) -> AuthNError:
        """
        Parse raw text into an appropriate AuthNError object.

        Args:
            r (requests.Response): The response from the AuthN cluster.

        Raises:
            AuthNError: If the error doesn't match any specific conditions.
            ErrClusterNotFound: If the error message indicates a missing cluster.
            ErrRoleNotFound: If the error message indicates a missing role.
            ErrUserNotFound: If the error message indicates a missing user.
            ErrRoleAlreadyExists: If the error message indicates a role already exists.
            ErrUserAlreadyExists: If the error message indicates a user already exists.
            ErrClusterAlreadyRegistered: If the error message indicates the cluster is already registered.
            ErrUserInvalidCredentials: If the error message indicates invalid user credentials.
        """
        exc_class = self.exc_class
        status, message, req_url = r.status_code, r.text, r.request.url

        if status == 401:
            if "invalid credentials" in message:
                exc_class = ErrUserInvalidCredentials
        if status == 404:
            if "cluster " in message:
                exc_class = ErrClusterNotFound
            if "role " in message:
                exc_class = ErrRoleNotFound
            if "user " in message:
                exc_class = ErrUserNotFound
        if status == 409:
            if "cluster " in message:
                exc_class = ErrClusterAlreadyRegistered
            if "role " in message:
                exc_class = ErrRoleAlreadyExists
            if "user " in message:
                exc_class = ErrUserAlreadyExists

        return exc_class(status, message, req_url or "")



#
# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
#

from aistore.sdk.request_client import RequestClient
from aistore.sdk.authn.types import TokenMsg
from aistore.sdk.utils import get_logger
from aistore.sdk.const import (
    HTTP_METHOD_DELETE,
    URL_PATH_AUTHN_TOKENS,
)

logger = get_logger(__name__)


class TokenManager:  # pylint: disable=duplicate-code
    """
    Manages token-related operations.

    This class provides methods to interact with tokens in the AuthN server.
    .

    Args:
        client (RequestClient): The RequestClient used to make HTTP requests.
    """

    def __init__(self, client: RequestClient):
        self._client = client

    @property
    def client(self) -> RequestClient:
        """Returns the RequestClient instance used by this TokenManager."""
        return self._client

    def revoke(self, token: str) -> None:
        """
        Revokes the specified authentication token.

        Args:
            token (str): The token to be revoked.

        Raises:
            ValueError: If the token is not provided.
            AISError: If the revoke token request fails.
        """
        if not token:
            raise ValueError("Token must be provided to revoke.")

        logger.info("Revoking token: %s", token)

        self.client.request(
            method=HTTP_METHOD_DELETE,
            path=f"{URL_PATH_AUTHN_TOKENS}",
            json=TokenMsg(token=token).dict(),
        )



#
# Copyright (c) 2024-2025, NVIDIA CORPORATION. All rights reserved.
#
from typing import List

from aistore.sdk.provider import Provider
from aistore.sdk.request_client import RequestClient
from aistore.sdk.authn.access_attr import AccessAttr
from aistore.sdk.authn.cluster_manager import ClusterManager
from aistore.sdk.types import BucketModel
from aistore.sdk.namespace import Namespace
from aistore.sdk.authn.errors import ErrRoleNotFound
from aistore.sdk.utils import get_logger
from aistore.sdk.const import (
    HTTP_METHOD_GET,
    HTTP_METHOD_POST,
    HTTP_METHOD_PUT,
    HTTP_METHOD_DELETE,
    URL_PATH_AUTHN_ROLES,
)
from aistore.sdk.authn.types import (
    RoleInfo,
    RolesList,
    BucketPermission,
    ClusterPermission,
)

logger = get_logger(__name__)


class RoleManager:
    """
    Manages role-related operations.

    This class provides methods to interact with roles, including
    retrieving, creating, updating, and deleting role information.

    Args:
        client (RequestClient): The RequestClient used to make HTTP requests.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, client: RequestClient):
        self._client = client

    @property
    def client(self) -> RequestClient:
        """Returns the RequestClient instance used by this RoleManager."""
        return self._client

    def list(self) -> RolesList:
        """
        Retrieves information about all roles.

        Returns:
            RoleList: A list containing information about all roles.

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore.
            requests.RequestException: If the HTTP request fails.
        """
        logger.info("Listing all roles")
        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=URL_PATH_AUTHN_ROLES,
            res_model=RolesList,
        )

    def get(self, role_name: str) -> RoleInfo:
        """
        Retrieves information about a specific role.

        Args:
            role_name (str): The name of the role to retrieve.

        Returns:
            RoleInfo: Information about the specified role.

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore.
            requests.RequestException: If the HTTP request fails.
        """
        logger.info("Getting role with name: %s", role_name)
        return self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=f"{URL_PATH_AUTHN_ROLES}/{role_name}",
            res_model=RoleInfo,
        )

    def create(
        self,
        name: str,
        desc: str,
        cluster_alias: str,
        perms: List[AccessAttr],
        bucket_name: str = None,
    ) -> RoleInfo:
        """
        Creates a new role.

        Args:
            name (str): The name of the role.
            desc (str): A description of the role.
            cluster_alias (str): The alias of the cluster this role will have access to.
            perms (List[AccessAttr]): A list of permissions to be granted for this role.
            bucket_name (str, optional): The name of the bucket this role will have access to.

        Returns:
            RoleInfo: Information about the newly created role.

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore.
            requests.RequestException: If the HTTP request fails.
        """
        # Convert the list of AccessAttr to an integer representing the permissions
        perm_value = sum(perm.value for perm in perms)

        cluster_uuid = ClusterManager(self.client).get(cluster_alias=cluster_alias).id
        role_info = RoleInfo(name=name, desc=desc)
        if bucket_name:
            role_info.buckets = [
                BucketPermission(
                    bck=BucketModel(
                        name=bucket_name,
                        provider=Provider.AIS.value,
                        namespace=Namespace(uuid=cluster_uuid),
                    ),
                    perm=perm_value,
                )
            ]
        else:
            role_info.clusters = [ClusterPermission(id=cluster_uuid, perm=perm_value)]

        logger.info("Creating role with name: %s", name)
        self.client.request(
            HTTP_METHOD_POST,
            path=URL_PATH_AUTHN_ROLES,
            json=role_info.dict(),
        )

        return self.get(role_name=name)

    def update(
        self,
        name: str,
        desc: str = None,
        cluster_alias: str = None,
        perms: List[AccessAttr] = None,
        bucket_name: str = None,
    ) -> RoleInfo:
        """
        Updates an existing role.

        Args:
            name (str): The name of the role.
            desc (str, optional): An updated description of the role.
            cluster_alias (str, optional): The alias of the cluster this role will have access to.
            perms (List[AccessAttr], optional): A list of updated permissions to be granted for this role.
            bucket_name (str, optional): The name of the bucket this role will have access to.

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore.
            requests.RequestException: If the HTTP request fails.
            ValueError: If the role does not exist or if invalid parameters are provided.
        """

        if not (desc or cluster_alias or perms or bucket_name):
            raise ValueError(
                "You must change either the description or permissions for a bucket or cluster."
            )

        if perms and not cluster_alias:
            raise ValueError(
                "Cluster alias must be provided when permissions are specified."
            )

        if bucket_name and not cluster_alias:
            raise ValueError(
                "Cluster alias must be provided when bucket_name is specified."
            )

        if (cluster_alias or bucket_name) and not perms:
            raise ValueError(
                "Permissions must be provided when cluster alias or bucket name is specified."
            )

        role_info = self.get(role_name=name)

        if desc:
            role_info.desc = desc

        if cluster_alias:
            cluster_uuid = (
                ClusterManager(self.client).get(cluster_alias=cluster_alias).id
            )
            perm_value = sum(perm.value for perm in perms)

            if bucket_name:
                logger.info(
                    "Preparing bucket-specific permissions for bucket: %s", bucket_name
                )
                role_info.buckets = [
                    BucketPermission(
                        bck=BucketModel(
                            name=bucket_name,
                            provider=Provider.AIS.value,
                            namespace=Namespace(uuid=cluster_uuid),
                        ),
                        perm=perm_value,
                    )
                ]
            else:
                logger.info(
                    "Preparing cluster-wide permissions for cluster alias: %s",
                    cluster_alias,
                )
                role_info.clusters = [
                    ClusterPermission(id=cluster_uuid, perm=perm_value)
                ]

        logger.info("Updating role with name: %s", name)
        self.client.request(
            HTTP_METHOD_PUT,
            path=f"{URL_PATH_AUTHN_ROLES}/{name}",
            json=role_info.dict(),
        )

    def delete(self, name: str, missing_ok: bool = False) -> None:
        """
        Deletes a role.

        Args:
            name (str): The name of the role to delete.
            missing_ok (bool): Ignore error if role does not exist. Defaults to False

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore.
            requests.RequestException: If the HTTP request fails.
            ValueError: If the role does not exist.
        """
        logger.info("Deleting role with name: %s", name)

        try:
            self.client.request(
                HTTP_METHOD_DELETE,
                path=f"{URL_PATH_AUTHN_ROLES}/{name}",
            )
        except ErrRoleNotFound as err:
            if not missing_ok:
                raise err



"""
Dynamic Batch Sampler for Dynamic Batch Sizing

In scenarios where memory is a constraint, the DynamicBatchSampler
can be used to generate mini-batches that fit within a memory constraint
so that there is a guarantee that each batch fits within memory
while attempting to fit the maximum number of samples in each batch.

Copyright (c) 2024-2025, NVIDIA CORPORATION. All rights reserved.
"""

from logging import getLogger
from typing import Iterator, List
import torch
from aistore.pytorch.base_map_dataset import AISBaseMapDataset

# Default saturation of a batch needed to not be dropped with drop_last=True
SATURATION_FACTOR = 0.8
logger = getLogger(__name__)


# pylint: disable=too-many-instance-attributes, too-many-arguments
class DynamicBatchSampler(torch.utils.data.Sampler):
    """

    Dynamically adds samples to mini-batch up to a maximum batch size.

    NOTE: Using this sampler with AISBaseMapDatasets that use ObjectGroups
    in their ais_source_lists will be slower than using it with Buckets as
    ObjectGroups will perform one extra API call per object to get size metadata.

    Args:
        data_source (AISBaseMapDataset): Base AIS map-style dataset to sample from to create dynamic mini-batches.
        max_batch_size (float): Maximum size of mini-batch in bytes.
        drop_last (bool, optional): If `True`, then will drop last batch if the batch is not atleast 80% of
            `max_batch_size`. Defaults to `False`.
        allow_oversized_samples (bool, optional): If `True`, then any sample that is larger than the `max_batch_size`
            will be processed in its own min-batch by itself instead of being dropped. Defaults to `False`.
        saturation_factor (float, optional): Saturation of a batch needed to not be dropped with `drop_last=True`.
            Default is `0.8`.
        shuffle (bool, optional): Randomizes order of samples before calculating mini-batches. Default is `False`.
    """

    def __init__(
        self,
        data_source: AISBaseMapDataset,
        max_batch_size: float,
        drop_last: bool = False,
        allow_oversized_samples: bool = False,
        saturation_factor: float = SATURATION_FACTOR,
        shuffle: bool = False,
    ) -> None:
        super().__init__()
        self._data_source = data_source
        self._max_batch_size = max_batch_size
        self._samples_list = data_source.get_obj_list()
        self._drop_last = drop_last
        self._allow_oversized_samples = allow_oversized_samples
        if not 0 <= saturation_factor <= 1:
            raise ValueError("`saturation_factor` must be between 0 and 1")
        self._saturation_factor = saturation_factor
        self._shuffle = shuffle
        self._indices = None

    def __iter__(self) -> Iterator[List[int]]:
        """
        Returns an iterator containing mini-batches (lists of indices).
        """
        total_mem = 0
        batch = []

        if self._shuffle:
            self._indices = list(
                torch.randperm(len(self))
            )  # randomized list of indices

        # Get sample size for each index, check if there is space in the batch, and yield batches whenever full
        # Calculate spaces in batch non-preemptively
        index = self._get_next_index(-1)
        while index < len(self._samples_list):
            sample = self._samples_list[index]

            if sample.props_cached.size == 0:
                logger.warning(
                    "Sample %s cannot be processed as it has a size of 0 bytes",
                    sample.name,
                )
                index = self._get_next_index(index)
                continue

            if sample.props_cached.size > self._max_batch_size:
                if self._allow_oversized_samples is True:
                    yield [index]
                else:
                    logger.warning(
                        "Sample %s cannot be processed as it is larger than the max batch size: %d bytes > %d bytes",
                        sample.name,
                        sample.props_cached.size,
                        self._max_batch_size,
                    )

                index = self._get_next_index(index)
                continue

            if total_mem + sample.props_cached.size < self._max_batch_size:
                batch.append(index)
                index = self._get_next_index(index)
                total_mem += sample.props_cached.size
            else:

                if total_mem + sample.props_cached.size == self._max_batch_size:
                    batch.append(index)
                    index = self._get_next_index(index)

                yield batch
                batch = []
                total_mem = 0

        # If batch exists and we are not dropping last or if we are dropping last but the batch is saturated
        # then yield the last batch
        if (batch and not self._drop_last) or (
            self._drop_last
            and total_mem / self._max_batch_size > self._saturation_factor
        ):
            yield batch

    def __len__(self) -> int:
        """
        Returns the total number of samples.
        """
        return len(self._samples_list)

    def _get_next_index(self, index) -> int:
        """
        Get next index from indices if shuffling or otherwise return incremented count.

        Returns:
            index (int): Next index to sample from
        """
        if self._shuffle:
            if len(self._indices) == 0:
                return len(self._samples_list)
            return int(self._indices.pop())
        index += 1
        return index



"""
Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
"""

import logging
from dataclasses import dataclass
from urllib3.util.retry import Retry
from tenacity import (
    Retrying,
    wait_exponential,
    stop_after_delay,
    retry_if_exception_type,
    before_sleep_log,
)
from requests.exceptions import (
    ConnectTimeout,
    ReadTimeout,
    ChunkedEncodingError,
    ConnectionError as RequestsConnectionError,
)
from aistore.sdk.errors import AISRetryableError

# Default Retry Exceptions
NETWORK_RETRY_EXCEPTIONS = (
    ConnectTimeout,
    ReadTimeout,
    ChunkedEncodingError,
    RequestsConnectionError,
    AISRetryableError,
)


@dataclass
class RetryConfig:
    """
    Configuration class for managing both HTTP and network retries in AIStore.

    AIStore implements two types of retries to ensure reliability and fault tolerance:

    1. **HTTP Retry (urllib3.Retry)** - Handles HTTP errors based on status codes (e.g., 429, 500, 502, 503, 504).
    2. **Network Retry (tenacity)** - Recovers from connection failures, timeouts, and unreachable targets.

    **Why two types of retries?**
    - AIStore uses **redirects** for GET/PUT operations.
    - If a target node is down, we must retry the request via the proxy instead of the same failing target.
    - `network_retry` ensures that the request is reattempted at the **proxy level**, preventing unnecessary failures.

    **Attributes:**
        http_retry (urllib3.Retry): Defines retry behavior for transient HTTP errors.
        network_retry (tenacity.Retrying): Configured `tenacity.Retrying` instance managing retries for network-related
            issues, such as connection failures, timeouts, or unreachable targets.
    """

    http_retry: Retry
    network_retry: Retrying

    @staticmethod
    def default() -> "RetryConfig":
        """
        Returns the default retry configuration for AIStore.
        """
        return RetryConfig(
            http_retry=Retry(
                total=3,
                backoff_factor=0.5,
                # NOTE: Status codes 429 (Too Many Requests) and 503 (Service Unavailable)
                # are temporarily handled here to mitigate rate-limiting issues.
                # This logic is intended for AIStore versions prior to 3.28 and should be removed afterward.
                # See: https://aistore.nvidia.com/blog/2025/03/19/rate-limit-blog
                # TODO (post-3.28):
                # Distinguish whether the rate limit occurred at the frontend (client-facing) or backend (cloud-facing).
                # - For backend rate-limiting, find and apply a better-suited config.
                #   See (`$ ais config cluster rate_limit --json`).
                # - For frontend rate-limiting, we need a smarter handling mechanism.
                #   The current HTTP retry strategy makes 3 attempts with 0.5s backoff,
                #   which may not be appropriate in all scenarios.
                status_forcelist=[429, 500, 502, 503, 504],
                connect=0,
                read=0,
            ),
            network_retry=Retrying(
                wait=wait_exponential(multiplier=1, min=1, max=10),
                stop=stop_after_delay(60),
                retry=retry_if_exception_type(NETWORK_RETRY_EXCEPTIONS),
                before_sleep=before_sleep_log(logging.getLogger(), logging.DEBUG),
                reraise=True,
            ),
        )



#
# Copyright (c) 2022-2025, NVIDIA CORPORATION. All rights reserved.
#

from __future__ import annotations

import json
import logging
import os
from pathlib import Path
import time
from typing import Dict, List, NewType, Iterable, Union, Optional
import requests
from requests import structures

from aistore.sdk.ais_source import AISSource
from aistore.sdk.etl.etl_const import DEFAULT_ETL_TIMEOUT
from aistore.sdk.obj.object_iterator import ObjectIterator
from aistore.sdk.etl import ETLConfig
from aistore.sdk.const import (
    ACT_COPY_BCK,
    ACT_CREATE_BCK,
    ACT_DESTROY_BCK,
    ACT_ETL_BCK,
    ACT_EVICT_REMOTE_BCK,
    ACT_LIST,
    ACT_MOVE_BCK,
    ACT_SUMMARY_BCK,
    HEADER_ACCEPT,
    HEADER_BUCKET_PROPS,
    HEADER_BUCKET_SUMM,
    HEADER_XACTION_ID,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_GET,
    HTTP_METHOD_HEAD,
    HTTP_METHOD_POST,
    MSGPACK_CONTENT_TYPE,
    QPARAM_BCK_TO,
    QPARAM_BSUMM_REMOTE,
    QPARAM_FLT_PRESENCE,
    QPARAM_KEEP_REMOTE,
    QPARAM_NAMESPACE,
    QPARAM_PROVIDER,
    QPARAM_UUID,
    URL_PATH_BUCKETS,
    STATUS_ACCEPTED,
    STATUS_OK,
    STATUS_PARTIAL_CONTENT,
    DEFAULT_JOB_POLL_TIME,
)
from aistore.sdk.enums import FLTPresence
from aistore.sdk.provider import Provider
from aistore.sdk.dataset.dataset_config import DatasetConfig

from aistore.sdk.errors import (
    InvalidBckProvider,
    ErrBckAlreadyExists,
    ErrBckNotFound,
    UnexpectedHTTPStatusCode,
)
from aistore.sdk.multiobj import ObjectGroup, ObjectRange
from aistore.sdk.request_client import RequestClient
from aistore.sdk.obj.object import Object, BucketDetails
from aistore.sdk.types import (
    ActionMsg,
    BucketEntry,
    BucketList,
    BucketModel,
    BsummCtrlMsg,
    Namespace,
    CopyBckMsg,
    TransformBckMsg,
    TCBckMsg,
    ListObjectsMsg,
)
from aistore.sdk.list_object_flag import ListObjectFlag
from aistore.sdk.utils import validate_directory, get_file_size
from aistore.sdk.obj.object_props import ObjectProps

Header = NewType("Header", structures.CaseInsensitiveDict)


# pylint: disable=too-many-public-methods,too-many-lines
class Bucket(AISSource):
    """
    A class representing a bucket that contains user data.

    Args:
        client (RequestClient): Client for interfacing with AIS cluster
        name (str): name of bucket
        provider (str or Provider, optional): Provider of bucket (one of "ais", "aws", "gcp", ...), defaults to "ais"
        namespace (Namespace, optional): Namespace of bucket, defaults to None
    """

    def __init__(
        self,
        name: str,
        client: RequestClient = None,
        provider: Union[Provider, str] = Provider.AIS,
        namespace: Namespace = None,
    ):
        self._client = client
        self._name = name
        self._provider = Provider.parse(provider)
        self._namespace = namespace
        self._qparam = {QPARAM_PROVIDER: self.provider.value}
        if self.namespace:
            self._qparam[QPARAM_NAMESPACE] = namespace.get_path()

    @property
    def client(self) -> RequestClient:
        """The client used by this bucket."""
        return self._client

    @client.setter
    def client(self, client):
        """Update the client used by this bucket."""
        self._client = client

    @property
    def qparam(self) -> Dict:
        """Default query parameters to use with API calls from this bucket."""
        return self._qparam

    @property
    def provider(self) -> Provider:
        """The provider for this bucket."""
        return self._provider

    @property
    def name(self) -> str:
        """The name of this bucket."""
        return self._name

    @property
    def namespace(self) -> Namespace:
        """The namespace for this bucket."""
        return self._namespace

    def list_urls(
        self, prefix: str = "", etl: Optional[ETLConfig] = None
    ) -> Iterable[str]:
        """
        Generates full URLs for all objects in the bucket that match the specified prefix.

        Args:
            prefix (str, optional): A string prefix to filter objects. Only objects with names starting
                with this prefix will be included. Defaults to an empty string (no filtering).
            etl (Optional[ETLConfig], optional): An optional ETL configuration. If provided, the URLs
                will include ETL processing parameters. Defaults to None.

        Returns:
            Iterable[str]: An iterator yielding full URLs of all objects matching the prefix.
        """
        for entry in self.list_objects_iter(prefix=prefix, props="name"):
            yield self.object(entry.name).get_url(etl=etl)

    def list_all_objects_iter(
        self, prefix: str = "", props: str = "name,size"
    ) -> Iterable[Object]:
        """
        Implementation of the abstract method from AISSource that provides an iterator
        of all the objects in this bucket matching the specified prefix.

        Args:
            prefix (str, optional): Limit objects selected by a given string prefix
            props (str, optional): Comma-separated list of object properties to return. Default value is "name,size".
                Properties: "name", "size", "atime", "version", "checksum", "target_url", "copies".

        Returns:
            Iterator of all object URLs matching the prefix
        """
        for entry in self.list_objects_iter(prefix=prefix, props=props):
            yield self.object(entry.name, entry.generate_object_props())

    def create(self, exist_ok=False):
        """
        Creates a bucket in AIStore cluster.
        Can only create a bucket for AIS provider on localized cluster. Remote cloud buckets do not support creation.

        Args:
            exist_ok (bool, optional): Ignore error if the cluster already contains this bucket

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            aistore.sdk.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        self._verify_ais_bucket()
        try:
            self.make_request(HTTP_METHOD_POST, ACT_CREATE_BCK)
        except ErrBckAlreadyExists as err:
            if not exist_ok:
                raise err
        return self

    def delete(self, missing_ok=False):
        """
        Destroys bucket in AIStore cluster.
        In all cases removes both the bucket's content _and_ the bucket's metadata from the cluster.
        Note: AIS will _not_ call the remote backend provider to delete the corresponding Cloud bucket
        (iff the bucket in question is, in fact, a Cloud bucket).

        Args:
            missing_ok (bool, optional): Ignore error if bucket does not exist

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            aistore.sdk.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        self._verify_ais_bucket()
        try:
            self.make_request(HTTP_METHOD_DELETE, ACT_DESTROY_BCK)
        except ErrBckNotFound as err:
            if not missing_ok:
                raise err

    def rename(self, to_bck_name: str) -> str:
        """
        Renames bucket in AIStore cluster.
        Only works on AIS buckets. Returns job ID that can be used later to check the status of the asynchronous
            operation.

        Args:
            to_bck_name (str): New bucket name for bucket to be renamed as

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            aistore.sdk.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        self._verify_ais_bucket()
        params = self.qparam.copy()
        params[QPARAM_BCK_TO] = Bucket(
            name=to_bck_name, namespace=self.namespace
        ).get_path()
        resp = self.make_request(HTTP_METHOD_POST, ACT_MOVE_BCK, params=params)
        self._name = to_bck_name
        return resp.text

    def evict(self, keep_md: bool = False):
        """
        Evicts bucket in AIStore cluster.
        NOTE: only Cloud buckets can be evicted.

        Args:
            keep_md (bool, optional): If true, evicts objects but keeps the bucket's metadata (i.e., the bucket's name
                and its properties)

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            aistore.sdk.errors.InvalidBckProvider: Invalid bucket provider for requested operation
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        self.verify_cloud_bucket()
        params = self.qparam.copy()
        params[QPARAM_KEEP_REMOTE] = str(keep_md)
        self.make_request(HTTP_METHOD_DELETE, ACT_EVICT_REMOTE_BCK, params=params)

    def head(self) -> Header:
        """
        Requests bucket properties.

        Returns:
            Response header with the bucket properties

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        return self.client.request(
            HTTP_METHOD_HEAD,
            path=f"{URL_PATH_BUCKETS}/{self.name}",
            params=self.qparam,
        ).headers

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def summary(
        self,
        uuid: str = "",
        prefix: str = "",
        cached: bool = True,
        present: bool = True,
    ):
        """
        Returns bucket summary (starts xaction job and polls for results).

        Args:
            uuid (str): Identifier for the bucket summary. Defaults to an empty string.
            prefix (str): Prefix for objects to be included in the bucket summary.
                          Defaults to an empty string (all objects).
            cached (bool): If True, summary entails cached entities. Defaults to True.
            present (bool): If True, summary entails present entities. Defaults to True.

        Raises:
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
            aistore.sdk.errors.AISError: All other types of errors with AIStore
        """
        bsumm_ctrl_msg = BsummCtrlMsg(
            uuid=uuid, prefix=prefix, cached=cached, present=present
        )

        # Start the job and get the job ID
        resp = self.make_request(
            HTTP_METHOD_GET,
            ACT_SUMMARY_BCK,
            params=self.qparam,
            value=bsumm_ctrl_msg.dict(),
        )

        # Initial response status code should be 202
        if resp.status_code == STATUS_OK:
            raise UnexpectedHTTPStatusCode([STATUS_ACCEPTED], resp.status_code)

        job_id = resp.text.strip('"')

        # Update the uuid in the control message
        bsumm_ctrl_msg.uuid = job_id

        # Sleep and request frequency in sec (starts at 0.2s)
        sleep_time = DEFAULT_JOB_POLL_TIME

        # Poll async task for http.StatusOK completion
        while True:
            resp = self.make_request(
                HTTP_METHOD_GET,
                ACT_SUMMARY_BCK,
                params=self.qparam,
                value=bsumm_ctrl_msg.dict(),
            )

            # If task completed successfully, break the loop
            if resp.status_code == STATUS_OK:
                break
            # If status code received is neither STATUS_ACCEPTED nor STATUS_PARTIAL_CONTENT, raise an exception
            if resp.status_code not in (STATUS_ACCEPTED, STATUS_PARTIAL_CONTENT):
                raise UnexpectedHTTPStatusCode(
                    [STATUS_OK, STATUS_ACCEPTED, STATUS_PARTIAL_CONTENT],
                    resp.status_code,
                )

            # If task is still running, wait for some time and try again
            time.sleep(sleep_time)
            if resp.status_code != STATUS_PARTIAL_CONTENT:
                sleep_time = min(
                    10, sleep_time * 1.5
                )  # Increase sleep_time by 50%, but don't exceed 10 seconds

        return json.loads(resp.content.decode("utf-8"))[0]

    def info(
        self,
        flt_presence: int = FLTPresence.FLT_EXISTS,
        bsumm_remote: bool = True,
        prefix: str = "",
    ):
        """
        Returns bucket summary and information/properties.

        Args:
            flt_presence (FLTPresence): Describes the presence of buckets and objects with respect to their existence
                                or non-existence in the AIS cluster using the enum FLTPresence. Defaults to
                                value FLT_EXISTS and values are:
                                FLT_EXISTS - object or bucket exists inside and/or outside cluster
                                FLT_EXISTS_NO_PROPS - same as FLT_EXISTS but no need to return summary
                                FLT_PRESENT - bucket is present or object is present and properly
                                located
                                FLT_PRESENT_NO_PROPS - same as FLT_PRESENT but no need to return summary
                                FLT_PRESENT_CLUSTER - objects present anywhere/how in
                                the cluster as replica, ec-slices, misplaced
                                FLT_EXISTS_OUTSIDE - not present; exists outside cluster
            bsumm_remote (bool): If True, returned bucket info will include remote objects as well
            prefix (str): Only include objects with the given prefix in the bucket

        Raises:
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
            ValueError: `flt_presence` is not one of the expected values
            aistore.sdk.errors.AISError: All other types of errors with AIStore
        """
        try:
            flt_presence = int(FLTPresence(flt_presence))
        except Exception as err:
            raise ValueError(
                "`flt_presence` must be in values of enum FLTPresence"
            ) from err

        params = self.qparam.copy()
        params.update({QPARAM_FLT_PRESENCE: flt_presence})
        params[QPARAM_BSUMM_REMOTE] = bsumm_remote

        path = f"{URL_PATH_BUCKETS}/{self.name}"
        if prefix:
            path += f"/{prefix}"

        response = self.client.request(
            HTTP_METHOD_HEAD,
            path=path,
            params=params,
        )

        bucket_props = response.headers.get(HEADER_BUCKET_PROPS, "{}")
        uuid = response.headers.get(HEADER_XACTION_ID, "").strip('"')
        params[QPARAM_UUID] = uuid

        result = {}

        # Initial response status code should be 202
        if response.status_code != int(STATUS_ACCEPTED):
            raise UnexpectedHTTPStatusCode([STATUS_ACCEPTED], response.status_code)

        # Sleep and request frequency in sec (starts at 2 s)
        sleep_time = 2
        time.sleep(sleep_time)
        i = 0

        # Poll async task for http.StatusOK completion
        while True:
            response = self.client.request(
                HTTP_METHOD_HEAD,
                path=f"{URL_PATH_BUCKETS}/{self.name}",
                params=params,
            )

            bucket_summ = response.headers.get(HEADER_BUCKET_SUMM, "")

            if bucket_summ != "":
                result = json.loads(bucket_summ)

            # If task completed successfully, break the loop
            if response.status_code == STATUS_OK:
                break

            time.sleep(sleep_time)
            i += 1

            if i == 8 and response.status_code != STATUS_PARTIAL_CONTENT:
                sleep_time *= 2
            elif i == 16 and response.status_code != STATUS_PARTIAL_CONTENT:
                sleep_time *= 2

        return bucket_props, result

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def copy(
        self,
        to_bck: Bucket,
        prefix_filter: str = "",
        prepend: str = "",
        ext: Optional[Dict[str, str]] = None,
        dry_run: bool = False,
        force: bool = False,
        latest: bool = False,
        sync: bool = False,
        num_workers: Optional[int] = 0,
    ) -> str:
        """
        Returns job ID that can be used later to check the status of the asynchronous operation.

        Args:
            to_bck (Bucket): Destination bucket
            prefix_filter (str, optional): Only copy objects with names starting with this prefix
            prepend (str, optional): Value to prepend to the name of copied objects
            ext (Dict[str, str], optional): Dict mapping each extension to the extension that will replace it
                (e.g. {"jpg": "txt"})
            dry_run (bool, optional): Determines if the copy should actually
                happen or not
            force (bool, optional): Override existing destination bucket
            latest (bool, optional): GET the latest object version from the associated remote bucket
            sync (bool, optional): synchronize destination bucket with its remote (e.g., Cloud or remote AIS) source
            num_workers (int, optional): Number of concurrent workers for the copy job per target
                - 0 (default): number of mountpaths
                - -1: single thread, serial execution

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        value = TCBckMsg(
            ext=ext,
            num_workers=num_workers,
            copy_msg=CopyBckMsg(
                prefix=prefix_filter,
                prepend=prepend,
                force=force,
                dry_run=dry_run,
                latest=latest,
                sync=sync,
            ),
        ).as_dict()

        params = self.qparam.copy()
        params[QPARAM_BCK_TO] = to_bck.get_path()
        return self.make_request(
            HTTP_METHOD_POST, ACT_COPY_BCK, value=value, params=params
        ).text

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def list_objects(
        self,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
        uuid: str = "",
        continuation_token: str = "",
        flags: List[ListObjectFlag] = None,
        target: str = "",
    ) -> BucketList:
        """
        Returns a structure that contains a page of objects, job ID, and continuation token (to read the next page, if
            available).

        Args:
            prefix (str, optional): Return only objects that start with the prefix
            props (str, optional): Comma-separated list of object properties to return. Default value is "name,size".
                Properties: "name", "size", "atime", "version", "checksum", "cached", "target_url", "status", "copies",
                "ec", "custom", "node".
            page_size (int, optional): Return at most "page_size" objects.
                The maximum number of objects in response depends on the bucket backend. E.g, AWS bucket cannot return
                    more than 5,000 objects in a single page.
                NOTE: If "page_size" is greater than a backend maximum, the backend maximum objects are returned.
                Defaults to "0" - return maximum number of objects.
            uuid (str, optional): Job ID, required to get the next page of objects
            continuation_token (str, optional): Marks the object to start reading the next page
            flags (List[ListObjectFlag], optional): Optional list of ListObjectFlag enums to include as flags in the
             request
            target(str, optional): Only list objects on this specific target node

        Returns:
            BucketList: the page of objects in the bucket and the continuation token to get the next page
            Empty continuation token marks the final page of the object list

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """

        value = ListObjectsMsg(
            prefix=prefix,
            page_size=page_size,
            uuid=uuid,
            props=props,
            continuation_token=continuation_token,
            flags=[] if flags is None else flags,
            target=target,
        ).as_dict()
        action = ActionMsg(action=ACT_LIST, value=value).dict()

        bucket_list = self.client.request_deserialize(
            HTTP_METHOD_GET,
            path=f"{URL_PATH_BUCKETS}/{ self.name }",
            headers={HEADER_ACCEPT: MSGPACK_CONTENT_TYPE},
            res_model=BucketList,
            json=action,
            params=self.qparam,
        )

        for entry in bucket_list.entries:
            entry.object = self.object(entry.name)
        return bucket_list

    def list_objects_iter(
        self,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
        flags: List[ListObjectFlag] = None,
        target: str = "",
    ) -> ObjectIterator:
        """
        Returns an iterator for all objects in bucket

        Args:
            prefix (str, optional): Return only objects that start with the prefix
            props (str, optional): Comma-separated list of object properties to return. Default value is "name,size".
                Properties: "name", "size", "atime", "version", "checksum", "cached", "target_url", "status", "copies",
                "ec", "custom", "node".
            page_size (int, optional): return at most "page_size" objects
                The maximum number of objects in response depends on the bucket backend. E.g, AWS bucket cannot return
                    more than 5,000 objects in a single page.
                NOTE: If "page_size" is greater than a backend maximum, the backend maximum objects are returned.
                Defaults to "0" - return maximum number objects
            flags (List[ListObjectFlag], optional): Optional list of ListObjectFlag enums to include as flags in the
             request
            target(str, optional): Only list objects on this specific target node

        Returns:
            ObjectIterator: object iterator

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """

        def fetch_objects(uuid, token):
            return self.list_objects(
                prefix,
                props,
                page_size,
                uuid=uuid,
                continuation_token=token,
                flags=flags,
                target=target,
            )

        return ObjectIterator(fetch_objects)

    def list_all_objects(
        self,
        prefix: str = "",
        props: str = "",
        page_size: int = 0,
        flags: List[ListObjectFlag] = None,
        target: str = "",
    ) -> List[BucketEntry]:
        """
        Returns a list of all objects in bucket

        Args:
            prefix (str, optional): return only objects that start with the prefix
            props (str, optional): comma-separated list of object properties to return. Default value is "name,size".
                Properties: "name", "size", "atime", "version", "checksum", "cached", "target_url", "status", "copies",
                "ec", "custom", "node".
            page_size (int, optional): return at most "page_size" objects
                The maximum number of objects in response depends on the bucket backend. E.g, AWS bucket cannot return
                    more than 5,000 objects in a single page.
                NOTE: If "page_size" is greater than a backend maximum, the backend maximum objects are returned.
                Defaults to "0" - return maximum number objects
            flags (List[ListObjectFlag], optional): Optional list of ListObjectFlag enums to include as flags in the
             request
            target(str, optional): Only list objects on this specific target node

        Returns:
            List[BucketEntry]: list of objects in bucket

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ReadTimeout: Timed out receiving response from AIStore
        """
        uuid = ""
        continuation_token = ""
        obj_list = None

        while True:
            resp = self.list_objects(
                prefix=prefix,
                props=props,
                page_size=page_size,
                uuid=uuid,
                continuation_token=continuation_token,
                flags=flags,
                target=target,
            )
            if obj_list:
                obj_list = obj_list + resp.entries
            obj_list = obj_list or resp.entries
            if resp.continuation_token == "":
                break
            continuation_token = resp.continuation_token
            uuid = resp.uuid
        return obj_list

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def transform(
        self,
        etl_name: str,
        to_bck: Bucket,
        timeout: str = DEFAULT_ETL_TIMEOUT,
        prefix_filter: str = "",
        prepend: str = "",
        ext: Optional[Dict[str, str]] = None,
        force: bool = False,
        dry_run: bool = False,
        latest: bool = False,
        sync: bool = False,
        num_workers: Optional[int] = 0,
    ) -> str:
        """
        Visits all selected objects in the source bucket and for each object, puts the transformed
        result to the destination bucket

        Args:
            etl_name (str): name of etl to be used for transformations
            to_bck (str): destination bucket for transformations
            timeout (str, optional): Timeout of the ETL job (e.g. 5m for 5 minutes)
            prefix_filter (str, optional): Only transform objects with names starting with this prefix
            prepend (str, optional): Value to prepend to the name of resulting transformed objects
            ext (Dict[str, str], optional): Dict mapping each extension to the extension that will replace it
                (e.g. {"jpg": "txt"})
            dry_run (bool, optional): determines if the copy should actually happen or not
            force (bool, optional): override existing destination bucket
            latest (bool, optional): GET the latest object version from the associated remote bucket
            sync (bool, optional): synchronize destination bucket with its remote (e.g., Cloud or remote AIS) source
            num_workers (int, optional): Number of concurrent workers for the transformation job per target
                - 0 (default): number of mountpaths
                - -1: single thread, serial execution

        Returns:
            Job ID (as str) that can be used to check the status of the operation
        """
        value = TCBckMsg(
            ext=ext,
            num_workers=num_workers,
            transform_msg=TransformBckMsg(etl_name=etl_name, timeout=timeout),
            copy_msg=CopyBckMsg(
                prefix=prefix_filter,
                prepend=prepend,
                force=force,
                dry_run=dry_run,
                latest=latest,
                sync=sync,
            ),
        ).as_dict()

        params = self.qparam.copy()
        params[QPARAM_BCK_TO] = to_bck.get_path()
        return self.make_request(
            HTTP_METHOD_POST, ACT_ETL_BCK, value=value, params=params
        ).text

    def put_files(
        self,
        path: str,
        prefix_filter: str = "",
        pattern: str = "*",
        basename: bool = False,
        prepend: str = None,
        recursive: bool = False,
        dry_run: bool = False,
        verbose: bool = True,
    ) -> List[str]:
        """
        Puts files found in a given filepath as objects to a bucket in AIS storage.

        Args:
            path (str): Local filepath, can be relative or absolute
            prefix_filter (str, optional): Only put files with names starting with this prefix
            pattern (str, optional): Shell-style wildcard pattern to filter files
            basename (bool, optional): Whether to use the file names only as object names and omit the path information
            prepend (str, optional): Optional string to use as a prefix in the object name for all objects uploaded
                No delimiter ("/", "-", etc.) is automatically applied between the prepend value and the object name
            recursive (bool, optional): Whether to recurse through the provided path directories
            dry_run (bool, optional): Option to only show expected behavior without an actual put operation
            verbose (bool, optional): Whether to print upload info to standard output

        Returns:
            List of object names put to a bucket in AIS

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            ValueError: The path provided is not a valid directory
        """
        validate_directory(path)
        file_iterator = (
            Path(path).rglob(pattern) if recursive else Path(path).glob(pattern)
        )
        obj_names = []
        dry_run_prefix = "Dry-run enabled. Proposed action:" if dry_run else ""

        logger = logging.getLogger(f"{__name__}.put_files")
        logger.disabled = not verbose
        for file in file_iterator:
            if not file.is_file() or not str(file.name).startswith(prefix_filter):
                continue
            obj_name = self._get_uploaded_obj_name(file, path, basename, prepend)
            if not dry_run:
                self.object(obj_name).get_writer().put_file(str(file))
            logger.info(
                "%s File '%s' uploaded as object '%s' with size %s",
                dry_run_prefix,
                file,
                obj_name,
                get_file_size(file),
            )
            obj_names.append(obj_name)
        logger.info(
            "%s Specified files from %s uploaded to bucket %s",
            dry_run_prefix,
            path,
            f"{self.provider}://{self.name}",
        )
        return obj_names

    @staticmethod
    def _get_uploaded_obj_name(file, root_path, basename, prepend):
        obj_name = str(file.relative_to(root_path)) if not basename else file.name
        if prepend:
            return prepend + obj_name
        return obj_name

    def object(
        self,
        obj_name: str,
        props: ObjectProps = None,
    ) -> Object:
        """
        Factory constructor for an object in this bucket.
        Does not make any HTTP request, only instantiates an object in a bucket owned by the client.

        Args:
            obj_name (str): Name of object
            props (ObjectProps, optional): Properties of the object, as updated by head(), optionally pre-initialized.

        Returns:
            The object created.
        """
        details = BucketDetails(self.name, self.provider, self.qparam, self.get_path())
        return Object(
            client=self.client,
            bck_details=details,
            name=obj_name,
            props=props,
        )

    def objects(
        self,
        obj_names: List = None,
        obj_range: ObjectRange = None,
        obj_template: str = None,
    ) -> ObjectGroup:
        """
        Factory constructor for multiple objects belonging to this bucket.

        Args:
            obj_names (list): Names of objects to include in the group
            obj_range (ObjectRange): Range of objects to include in the group
            obj_template (str): String template defining objects to include in the group

        Returns:
            The ObjectGroup created
        """
        return ObjectGroup(
            bck=self,
            obj_names=obj_names,
            obj_range=obj_range,
            obj_template=obj_template,
        )

    def make_request(
        self,
        method: str,
        action: str,
        value: Dict = None,
        params: Dict = None,
    ) -> requests.Response:
        """
        Use the bucket's client to make a request to the bucket endpoint on the AIS server

        Args:
            method (str): HTTP method to use, e.g. POST/GET/DELETE
            action (str): Action string used to create an ActionMsg to pass to the server
            value (dict): Additional value parameter to pass in the ActionMsg
            params (dict, optional): Optional parameters to pass in the request

        Returns:
            Response from the server

        """
        if self._client is None:
            raise ValueError(
                "Bucket requires a client to use functions. Try defining a client and accessing this bucket with "
                "client.bucket()"
            )
        json_val = ActionMsg(action=action, value=value).dict()
        return self._client.request(
            method,
            path=f"{URL_PATH_BUCKETS}/{self.name}",
            json=json_val,
            params=params if params else self.qparam,
        )

    def _verify_ais_bucket(self):
        """
        Verify the bucket provider is AIS
        """
        if self.provider.is_remote():
            raise InvalidBckProvider(self.provider)

    def verify_cloud_bucket(self):
        """
        Verify the bucket provider is a cloud provider
        """
        if not self.provider.is_remote():
            raise InvalidBckProvider(self.provider)

    def get_path(self) -> str:
        """
        Get the path representation of this bucket
        """
        namespace_path = self.namespace.get_path() if self.namespace else "@#"
        return f"{ self.provider.value }/{ namespace_path }/{ self.name }/"

    def as_model(self) -> BucketModel:
        """
        Return a data-model of the bucket

        Returns:
            BucketModel representation
        """
        return BucketModel(
            name=self.name, namespace=self.namespace, provider=self.provider.value
        )

    def write_dataset(
        self,
        config: DatasetConfig,
        skip_missing: bool = True,
        **kwargs,
    ):
        """
        Write a dataset to a bucket in AIS in webdataset format using wds.ShardWriter. Logs the missing attributes

        Args:
            config (DatasetConfig): Configuration dict specifying how to process
                and store each part of the dataset item
            skip_missing (bool, optional): Skip samples that are missing one or more attributes, defaults to True
            **kwargs (optional): Optional keyword arguments to pass to the ShardWriter
        """

        # Add the upload shard logic to the original post-processing function
        original_post = kwargs.get("post", lambda path: None)

        def combined_post_processing(shard_path):
            original_post(shard_path)
            self.object(shard_path).get_writer().put_file(shard_path)
            os.unlink(shard_path)

        kwargs["post"] = combined_post_processing
        config.write_shards(skip_missing=skip_missing, **kwargs)



"""
Utils for AIS PyTorch Plugin

Copyright (c) 2022-2024, NVIDIA CORPORATION. All rights reserved.
"""

from urllib.parse import urlunparse
from typing import Tuple
from aistore.sdk.utils import parse_url as sdk_parse_url
from math import floor

MB_TO_B = 1000000


def convert_mb_to_bytes(megabytes: float) -> int:
    """
    Converts megabytes to bytes and truncates any extra bytes (floor).

    Args:
        megabytes (float): number of megabytes to convert

    Returns:
        int: number of bytes after conversion (floor of actual byte value)
    """
    return floor(megabytes * MB_TO_B)


def convert_bytes_to_mb(bytes: int) -> float:
    """
    Converts byes to megabytes.

    Args:
        bytes (int): number of bytes to convert to megabytes

    Returns:
        float: number of megabytes after conversion
    """
    return bytes / MB_TO_B


def unparse_url(provider: str, bck_name: str, obj_name: str) -> str:
    """
    Generate URL based on provider, bucket name, and object name.

    Args:
        provider (str): Provider name ('ais', 'gcp', etc.)
        bck_name (str): Bucket name
        obj_name (str): Object name with extension

    Returns:
        str: Complete URL
    """
    return urlunparse([provider, bck_name, obj_name, "", "", ""])


def get_basename(name: str) -> str:
    """
    Get the basename of the object name by stripping any directory information and suffix.

    Args:
        name (str): Complete object name

    Returns:
        str: Basename of the object
    """

    return name.split("/")[-1].split(".")[0]


def get_extension(name: str) -> str:
    """
    Get the file extension of the object by stripping any basename or prefix.

    Args:
        name (str): Complete object name

    Returns:
        str: File extension of the object
    """

    return name.split(".")[1]


def parse_url(url: str) -> Tuple[str, str, str]:
    """
    Wrapper of sdk/utils.py parse_url. Parse AIS URLs for bucket and object names.
    TODO: This can be removed once the upstream torch package for aiso is updated.

    Args:
        url (str): Complete URL of the object (e.g., "ais://bucket1/file.txt")

    Returns:
        Tuple[str, str, str]: Provider, bucket name, and object name
    """
    return sdk_parse_url(url)
